"""Defines a loader mixin to load xarray objects"""

import xarray as xr

from ..containers import PassthroughContainer, XrDataContainer
from ._tools import add_loader

# -----------------------------------------------------------------------------


class XarrayLoaderMixin:
    """Supplies functionality to load xarray objects"""

    @add_loader(TargetCls=XrDataContainer)
    def _load_xr_dataarray(
        filepath: str,
        *,
        TargetCls: type,
        load_completely: bool = False,
        **load_kwargs,
    ) -> XrDataContainer:
        """Loads an xr.DataArray from a netcdf file into an XrDataContainer.

        Args:
            filepath (str): Where the xarray-dumped netcdf file is located
            TargetCls (type): The class constructor
            load_completely (bool, optional): If true, will call .load() on the
                loaded DataArray to load it completely into memory
            **load_kwargs: Passed on to xr.load_dataarray, see there for kwargs

        Returns:
            XrDataContainer: The reconstructed XrDataContainer
        """
        da = xr.load_dataarray(filepath, **load_kwargs)

        if load_completely:
            da = da.load()

        # Create the XrDataContainer, carrying over attributes
        return TargetCls(data=da, attrs=da.attrs)

    @add_loader(TargetCls=PassthroughContainer)
    def _load_xr_dataset(
        filepath: str,
        *,
        TargetCls: type,
        load_completely: bool = False,
        **load_kwargs,
    ) -> PassthroughContainer:
        """Loads an xr.Dataset from a netcdf file into a PassthroughContainer.

        .. note::
            As there is no proper equivalent of a dataset in dantro (yet), and
            unpacking the dataset into a dantro group would reduce
            functionality, the PassthroughContainer is used here. It should
            behave almost the same as an xr.Dataset.

        Args:
            filepath (str): Where the xarray-dumped netcdf file is located
            TargetCls (type): The class constructor
            load_completely (bool, optional): If true, will call .load() on the
                loaded xr.Dataset to load it completely into memory.
            **load_kwargs: Passed on to xr.load_dataarray, see there for kwargs

        Returns:
            PassthroughContainer: The reconstructed XrDataset, stored in a
                passthrough container.
        """
        ds = xr.load_dataset(filepath, **load_kwargs)

        if load_completely:
            ds = ds.load()

        # Create the PassthroughContainer, carrying over dataset attributes
        return TargetCls(data=ds, attrs=ds.attrs)



"""Implements the Link class and the LinkContainer."""

import logging

from ..mixins import CheckDataMixin
from ..utils import Link
from .general import ObjectContainer, PassthroughContainer

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


class LinkContainer(CheckDataMixin, PassthroughContainer):
    """A LinkContainer is a container containing a Link object.

    It forwards all attribute calls to the Link object, which in turn forwards
    all attribute calls to the linked object, thereby emulating the behaviour
    of the linked object.
    """

    DATA_EXPECTED_TYPES = (Link,)

    def _format_info(self) -> str:
        """A __format__ helper function: returns info about the item

        In this case, the anchor and relative path of the associated link is
        returned.
        """
        return "{} -> {}, {}".format(
            self.data.anchor_object.name,
            self.data.target_rel_path,
            super(ObjectContainer, self)._format_info(),
        )



"""This module implements the ExternalPlotCreator class, which specialises on
creating matplotlib-based plots. These are accessed via 'external' modules
being imported and plot functions defined in these modules being invoked.
"""

import copy
import importlib
import importlib.util
import logging
import os
from typing import Callable, List, Sequence, Tuple, Union

import matplotlib as mpl

mpl.use("Agg")  # TODO Remove this. Should not be necessary!
import matplotlib.animation
import matplotlib.pyplot as plt

from ..dag import TransformationDAG
from ..tools import DoNothingContext, load_yml, recursive_update
from ._movie_writers import FileWriter
from ._plot_helper import (
    EnterAnimationMode,
    ExitAnimationMode,
    PlotHelper,
    PlotHelperError,
    PlotHelperErrors,
)
from .pcr_base import BasePlotCreator, _resolve_placeholders

# Local constants
log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------
# Tools


class figure_leak_prevention:
    """Context manager that aims to prevent superfluous matplotlib figures
    persisting beyond the context. Such figure objects can aggregate and start
    memory issues or even representation errors.

    Specifically, it does the following:

        * When entering, stores all current figure numbers
        * When exiting regularly, all figures that were opened within the
          context are closed, except the currently selected figure.
        * When exiting with an exception, the behaviour is the same, unless the
          ``close_current_fig_on_raise`` is set, in which case the currently
          selected figure is **not** excluded from closing.

    """

    def __init__(self, *, close_current_fig_on_raise: bool = False):
        """Initialize the context manager

        Args:
            close_current_fig_on_raise (bool, optional): If True, the
                currently selected figure will **not** be exempt from the
                figure closure in case an exception occurs. This flag has no
                effect when the context is exited without an exception.
        """
        self._fignums = None
        self._close_current = close_current_fig_on_raise

    def __enter__(self):
        """Upon entering, store all currently open figure numbers"""
        self._fignums = plt.get_fignums()
        log.trace(
            "Entering figure_leak_prevention context. Open figures: %s",
            self._fignums,
        )

    def __exit__(self, exc_type: type, *args) -> None:
        """Iterates over all currently open figures and closes all figures
        that were not previously open, except the currently selected figure.

        If an exception is detected, i.e. ``exc_type` is **not** None, the
        current figure is only closed if the context manager was entered with
        the ``close_current_fig_on_raise`` flag set.
        """
        log.trace(
            "Exiting figure_leak_prevention (exception: %s) ...", exc_type
        )

        # Determine whether to exclude the current figure or not
        exclude_current = not self._close_current or exc_type is None
        cfn = plt.gcf().number
        log.trace("  Current figure: %d", cfn)

        for n in plt.get_fignums():
            if n in self._fignums or (exclude_current and n == cfn):
                continue
            plt.close(n)
            log.trace("  Closed figure %d.", n)


# -----------------------------------------------------------------------------


class ExternalPlotCreator(BasePlotCreator):
    """This PlotCreator uses external scripts to create plots."""

    # Settings of functionality implemented in parent classes
    EXTENSIONS = "all"  # no checks performed
    DEFAULT_EXT = None
    DEFAULT_EXT_REQUIRED = False
    DAG_SUPPORTED = True
    DAG_INVOKE_IN_BASE = False  # False: DAG invocation NOT done automatically

    # For relative module imports, see the following as the base package
    BASE_PKG = "dantro.plot_creators.ext_funcs"

    # Which plot helper class to use
    PLOT_HELPER_CLS = PlotHelper

    # Configuration for the PlotManager's auto-detection feature ..............

    # Whether to ignore function attributes (e.g. `creator_name`) when
    # deciding whether a plot function is to be used with this creator
    _AD_IGNORE_FUNC_ATTRS = False

    # .........................................................................
    # Main API functions, required by PlotManager

    def __init__(
        self,
        name: str,
        *,
        base_module_file_dir: str = None,
        style: dict = None,
        **parent_kwargs,
    ):
        """Initialize an ExternalPlotCreator.

        Args:
            name (str): The name of this plot
            base_module_file_dir (str, optional): If given, ``module_file``
                arguments to the ``_plot`` method that are relative paths will
                be seen relative to this directory
            style (dict, optional): The default style context defintion to
                enter before calling the plot function. This can be used to
                specify the aesthetics of a plot. It is evaluated here once,
                stored as attribute, and can be updated when the plot method
                is called.
            **parent_kwargs: Passed to the parent __init__

        Raises:
            ValueError: On invalid ``base_module_file_dir`` argument
        """
        super().__init__(name, **parent_kwargs)

        # If given, check the base module file dir argument is valid
        if base_module_file_dir:
            bmfd = os.path.expanduser(base_module_file_dir)

            if not os.path.isabs(bmfd):
                raise ValueError(
                    "Argument `base_module_file_dir` needs to be "
                    f"an absolute path, was not! Got: {bmfd}"
                )

            elif not os.path.exists(bmfd) or not os.path.isdir(bmfd):
                raise ValueError(
                    "Argument `base_module_file_dir` does not "
                    "exists or does not point to a directory!"
                )

        self.base_module_file_dir = base_module_file_dir

        # Parse default RC parameters
        self._default_rc_params = None

        if style is not None:
            self._default_rc_params = self._prepare_style_context(**style)

    def plot(
        self,
        *,
        out_path: str,
        plot_func: Union[str, Callable],
        module: str = None,
        module_file: str = None,
        style: dict = None,
        helpers: dict = None,
        animation: dict = None,
        use_dag: bool = None,
        **func_kwargs,
    ):
        """Performs the plot operation by calling a specified plot function.

        The plot function is specified by its name, which is interpreted as a
        full module string, or by directly passing a callable.

        Alternatively, the base module can be loaded from a file path.

        Args:
            out_path (str): The output path for the resulting file
            plot_func (Union[str, Callable]): The plot function or a name or
                module string under which it can be imported.
            module (str, optional): If plot_func was the name of the plot
                function, this needs to be the name of the module to import
            module_file (str, optional): Path to the file to load and look for
                the ``plot_func`` in. If ``base_module_file_dir`` is given,
                this can also be a path relative to that directory.
            style (dict, optional): Parameters that determine the aesthetics of
                the created plot; basically matplotlib rcParams. From them, a
                style context is entered before calling the plot function.
                Valid keys:

                    base_style (str, List[str], optional):
                        names of valid matplotlib styles
                    rc_file (str, optional):
                        path to a YAML RC parameter file that is used to
                        update the base style
                    ignore_defaults (bool, optional):
                        Whether to ignore the default style passed to the
                        __init__ method
                    further keyword arguments:
                        will update the RC parameter dict yet again. Need be
                        valid matplotlib RC parameters in order to have any
                        effect.

            helpers (dict, optional): helper configuration passed to PlotHelper
                initialization if enabled
            animation (dict, optional): animation configuration
            use_dag (bool, optional): Whether to use the TransformationDAG to
                select and transform data that can be used in the plotting
                function. If not given, will query the plot function attributes
                for whether the DAG should be used.
            **func_kwargs: Passed to the imported function

        Raises:
            ValueError: On superfluous ``helpers`` or ``animation`` arguments
                in cases where these are not supported
        """
        # Get the plotting function
        plot_func = self._resolve_plot_func(
            plot_func=plot_func, module=module, module_file=module_file
        )

        # Generate a style dictionary to be used for context manager creation
        rc_params = self._prepare_style_context(**(style if style else {}))

        # Check if PlotHelper is to be used
        if getattr(plot_func, "use_helper", False):
            switch_anim_mode = False

            # Delegate to private helper method that performs the plot or the
            # animation. In case that animation mode is to be entered or
            # exited, adjust the animation-related parameters accordingly.
            try:
                self._plot_with_helper(
                    out_path=out_path,
                    plot_func=plot_func,
                    helpers=helpers,
                    style_context=self._build_style_context(**rc_params),
                    func_kwargs=func_kwargs,
                    use_dag=use_dag,
                    animation=animation,
                )

            except EnterAnimationMode:
                log.note("Entering animation mode ...")
                if not animation:
                    raise ValueError(
                        "Cannot dynamically enter animation mode without any "
                        "`animation` parameters having been specified in the "
                        f"configuration of the {self.classname} "
                        f"'{self.name}' plot!"
                    )

                switch_anim_mode = True
                animation = copy.deepcopy(animation)
                animation["enabled"] = True

            except ExitAnimationMode:
                log.note("Exiting animation mode ...")
                switch_anim_mode = True
                animation = None

            # else: animation was successful.

            # In case of the mode having switched, plot anew.
            if switch_anim_mode:
                log.debug("Plotting anew (with change in animation mode) ...")
                try:
                    self._plot_with_helper(
                        out_path=out_path,
                        plot_func=plot_func,
                        helpers=helpers,
                        style_context=self._build_style_context(**rc_params),
                        func_kwargs=func_kwargs,
                        use_dag=use_dag,
                        animation=animation,
                    )

                except (EnterAnimationMode, ExitAnimationMode):
                    raise RuntimeError(
                        "Cannot repeatedly enter or exit animation mode! Make "
                        f"sure that the plotting function of {self.logstr} "
                        "respects this requirement and that the plot "
                        "configuration you specified does not contradict "
                        "itself."
                    )

        else:
            # Call only the plot function
            # Do not allow helper or animation parameters
            if helpers:
                raise ValueError(
                    "The key 'helpers' was found in the configuration of "
                    f"plot '{self.name}' but usage of the PlotHelper is not "
                    f"supported by plot function '{plot_func.__name__}'!"
                )

            if animation:
                raise ValueError(
                    "The key 'animation' was found in the "
                    f"configuration of plot '{self.name}' but the animation "
                    "feature is only available when using the PlotHelper for "
                    f"plot function '{plot_func.__name__}'!"
                )

            # Prepare the arguments. The DataManager is added to args there
            # and data transformation via DAG occurs there as well.
            args, kwargs = self._prepare_plot_func_args(
                plot_func, use_dag=use_dag, out_path=out_path, **func_kwargs
            )

            # Enter the stlye context
            with self._build_style_context(**rc_params):
                log.debug(
                    "Calling plotting function '%s' ...", plot_func.__name__
                )
                plot_func(*args, **kwargs)
            # Done.

    def can_plot(self, creator_name: str, **cfg) -> bool:
        """Whether this plot creator is able to make a plot for the given plot
        configuration.

        This checks whether the configuration allows resolving a plot function.
        If that is the case, it checks whether the plot function has defined
        some attributes that provide further information on whether the current
        creator is the desired one.

        Args:
            creator_name (str): The name for this creator used within the
                PlotManager.
            **cfg: The plot configuration with which to decide this ...

        Returns:
            bool: Whether this creator can be used for plotting or not
        """
        log.debug(
            "Checking if %s can plot the given configuration ...", self.logstr
        )

        # Gather the arguments needed for plot function resolution and remove
        # those that are None
        pf_kwargs = dict(
            plot_func=cfg.get("plot_func"),
            module=cfg.get("module"),
            module_file=cfg.get("module_file"),
        )
        pf_kwargs = {k: v for k, v in pf_kwargs.items() if v is not None}

        # Try to resolve the function
        try:
            pf = self._resolve_plot_func(**pf_kwargs)

        except Exception:
            log.debug(
                "Cannot plot this configuration, because a plotting function "
                "could not be resolved with the given arguments: %s",
                pf_kwargs,
            )
            return False

        # else: was able to resolve a plotting function

        # The function might have an attribute that specifies the name of the
        # creator to use
        if not self._AD_IGNORE_FUNC_ATTRS:
            if self._declared_plot_func_by_attrs(pf, creator_name):
                return True

        # Nothing worked: This creator is not suitable.
        return False

    # .........................................................................
    # Helpers: Main plot routines

    def _plot_with_helper(
        self,
        *,
        out_path: str,
        plot_func: Callable,
        helpers: dict,
        style_context,
        func_kwargs: dict,
        animation: dict,
        use_dag: bool,
    ):
        """A helper method that performs plotting using the
        :py:class:`~dantro.plot_creators._plot_helper.PlotHelper`.

        Args:
            out_path (str): The output path
            plot_func (Callable): The resolved plot function
            helpers (dict): The helper configuration
            style_context: A style context; can also be DoNothingContext, if
                no style adjustments are to take place.
            func_kwargs (dict): Plot function arguments
            animation (dict): Animation parameters
            use_dag (bool): Whether a DAG is used in preprocessing or not
        """
        # Determine if animation is enabled, which is relevant for PlotHelper
        animation = copy.deepcopy(animation) if animation else {}
        animation_enabled = animation.pop("enabled", False)

        # Prepare the arguments. The DataManager is added to args there; if the
        # DAG is used, data transformation and placeholder resolution will
        # happen there as well.
        # In order to apply placeholder resolution to the helper configuration
        # as well, the helpers are passed along here (and popped from the
        # parsed kwargs again a few lines below).
        args, kwargs = self._prepare_plot_func_args(
            plot_func, use_dag=use_dag, helpers=helpers, **func_kwargs
        )

        # Initialize a PlotHelper instance that will take care of figure
        # setup, invoking helper-functions and saving the figure.
        # Then, add the Helper instance to the plot function keyword arguments.
        helpers = kwargs.pop("helpers")
        helper_defaults = getattr(plot_func, "helper_defaults", None)
        hlpr = self.PLOT_HELPER_CLS(
            out_path=out_path,
            helper_defaults=helper_defaults,
            update_helper_cfg=helpers,
            raise_on_error=self.raise_exc,
            animation_enabled=animation_enabled,
        )
        kwargs["hlpr"] = hlpr

        # Check if an animation is to be done; if so, delegate to helper method
        if animation_enabled:
            self._perform_animation(
                hlpr=hlpr,
                style_context=style_context,
                plot_func=plot_func,
                plot_args=args,
                plot_kwargs=kwargs,
                **animation,
            )
            return
        # else: No animation to be done.

        # Enter two context: one for style (could also be DoNothingContext)
        # and one for prevention of figures leaking from the plot function.
        leak_prev = figure_leak_prevention(close_current_fig_on_raise=True)

        with style_context, leak_prev:
            hlpr.setup_figure()

            log.info(
                "Now calling plotting function '%s' ...",
                plot_func.__name__,
            )
            plot_func(*args, **kwargs)

            log.note("Plot function finished. Finishing up ...")
            hlpr.invoke_enabled(axes="all")
            hlpr.save_figure()

    # .........................................................................
    # Helpers: Plot function resolution and argument preparation

    def _resolve_plot_func(
        self,
        *,
        plot_func: Union[str, Callable],
        module: str = None,
        module_file: str = None,
    ) -> Callable:
        """
        Args:
            plot_func (Union[str, Callable]): The plot function or a name or
                module string under which it can be imported.
            module (str): If plot_func was the name of the plot
                function, this needs to be the name of the module to import
            module_file (str): Path to the file to load and look for
                the `plot_func` in. If `base_module_file_dir` is given, this
                can also be a path relative to that directory.

        Returns:
            Callable: The resolved plot function

        Raises:
            TypeError: Upon wrong argument types
        """
        if callable(plot_func):
            log.debug("Received plotting function:  %s", str(plot_func))
            return plot_func

        elif not isinstance(plot_func, str):
            raise TypeError(
                "Argument `plot_func` needs to be a string or a "
                f"callable, was {type(plot_func)} with value '{plot_func}'."
            )

        # else: need to resolve the module and find the plot_func in it
        # First resolve the module, either from file or via import
        if module_file:
            mod = self._get_module_from_file(module_file)

        elif isinstance(module, str):
            mod = self._get_module_via_import(module)

        else:
            raise TypeError(
                "Could not import a module, because neither argument "
                "`module_file` was given nor did argument `module` have the "
                f"correct type (needs to be string but was {type(module)} "
                f"with value '{module}')."
            )

        # plot_func could be something like "A.B.C.d"; go along the segments to
        # allow for more versatile plot function retrieval
        attr_names = plot_func.split(".")
        for attr_name in attr_names[:-1]:
            mod = getattr(mod, attr_name)

        # This is now the last module. Get the actual function
        plot_func = getattr(mod, attr_names[-1])

        log.debug("Resolved plotting function:  %s", str(plot_func))

        return plot_func

    def _get_module_from_file(self, path: str):
        """Returns the module corresponding to the file at the given `path`"""
        # Pre-processing
        path = os.path.expanduser(path)

        # Make it absolute
        if not os.path.isabs(path):
            if not self.base_module_file_dir:
                raise ValueError(
                    "Need to specify `base_module_file_dir` "
                    "during initialization to use relative paths "
                    "for `module_file` argument!"
                )

            path = os.path.join(self.base_module_file_dir, path)

        # Extract a name from the path to use as module name
        mod_name = "from_file." + os.path.basename(path).split(".")[0]
        # NOTE The name does not really matter

        # Is an absolute path now
        # Create a module specification and, from that, import the module
        spec = importlib.util.spec_from_file_location(mod_name, path)
        mod = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(mod)

        # Now, it is loaded
        return mod

    def _get_module_via_import(self, module: str):
        """Returns the module via import"""
        # Import module via importlib, allowing relative imports from the
        # package defined as base package
        return importlib.import_module(module, package=self.BASE_PKG)

    def _prepare_plot_func_args(
        self, plot_func: Callable, *args, use_dag: bool = None, **kwargs
    ) -> Tuple[tuple, dict]:
        """Prepares the args and kwargs passed to the plot function.

        The passed args and kwargs are carried over, while the positional
        arguments are prepended with passing of the data manager.

        When subclassing this function, the parent method (this one) should
        still be called to maintain base functionality.

        Args:
            *args: Additional args
            **kwargs: Additional kwargs

        Returns:
            tuple: (args: tuple, kwargs: dict)
        """
        # If enabled, use the DAG interface to perform data selection. The
        # returned kwargs are the adjusted plot function keyword arguments.
        using_dag, kwargs = self._perform_data_selection(
            use_dag=use_dag, plot_kwargs=kwargs, _plot_func=plot_func
        )

        # Aggregate as (args, kwargs), passed on to plot function. When using
        # the DAG, the DataManager is NOT passed along, as it is accessible via
        # the tags of the DAG.
        if not using_dag:
            return ((self.dm,) + args, kwargs)
        return (args, kwargs)

    # .........................................................................
    # Helpers: specialization of data selection and transformation framework

    def _get_dag_params(
        self, *, _plot_func: Callable, **cfg
    ) -> Tuple[dict, dict]:
        """Extends the parent method by making the plot function callable
        available to the other helper methods and extracting some further
        information from the plot function.
        """
        dag_params, plot_kwargs = super()._get_dag_params(**cfg)

        # Store the plot function, such that it is available as argument in the
        # other subclassed helper methods
        dag_params["init"]["_plot_func"] = _plot_func
        dag_params["compute"]["_plot_func"] = _plot_func

        # Determine whether the DAG object should be passed along to the func
        pass_dag = getattr(_plot_func, "pass_dag_object_along", False)
        dag_params["pass_dag_object_along"] = pass_dag

        # Determine whether the DAG results should be unpacked when passing
        # them to the plot function
        unpack_results = getattr(_plot_func, "unpack_dag_results", False)
        dag_params["unpack_dag_results"] = unpack_results

        return dag_params, plot_kwargs

    def _use_dag(
        self, *, use_dag: bool, plot_kwargs: dict, _plot_func: Callable
    ) -> bool:
        """Whether the DAG should be used or not. This method extends that of
        the base class by additionally checking the plot function attributes
        for any information regarding the DAG
        """
        # If None was given, check the plot function attributes
        if use_dag is None:
            use_dag = getattr(_plot_func, "use_dag", None)

        # Let the parent class do whatever else it does
        use_dag = super()._use_dag(use_dag=use_dag, plot_kwargs=plot_kwargs)

        # Complain, if tags where required, but DAG usage was disabled
        if not use_dag and getattr(_plot_func, "required_dag_tags", None):
            raise ValueError(
                "The plot function {} requires DAG tags to be "
                "computed, but DAG usage was disabled."
                "".format(_plot_func)
            )

        return use_dag

    def _create_dag(
        self, *, _plot_func: Callable, **dag_params
    ) -> TransformationDAG:
        """Extends the parent method by allowing to pass the _plot_func, which
        can be used to adjust DAG behaviour ...
        """
        return super()._create_dag(**dag_params)

    def _compute_dag(
        self,
        dag: TransformationDAG,
        *,
        _plot_func: Callable,
        compute_only: Sequence[str],
        **compute_kwargs,
    ) -> dict:
        """Compute the dag results.

        This extends the parent method by additionally checking whether all
        required tags are defined and (after computation) whether all required
        tags were computed.
        """
        # Extract the required tags from the plot function
        required_tags = getattr(_plot_func, "required_dag_tags", None)

        # Make sure that all required tags are actually defined
        if required_tags:
            missing_tags = [t for t in required_tags if t not in dag.tags]

            if missing_tags:
                raise ValueError(
                    "Plot function {} required tags that were "
                    "not specified in the DAG: {}. Available "
                    "tags: {}. Please adjust the DAG "
                    "specification accordingly."
                    "".format(
                        _plot_func,
                        ", ".join(missing_tags),
                        ", ".join(dag.tags),
                    )
                )

        # If the compute_only argument was not explicitly given, determine
        # whether to compute only the required tags
        if (
            compute_only is None
            and required_tags is not None
            and getattr(_plot_func, "compute_only_required_dag_tags", False)
        ):
            log.remark(
                "Computing only tags that were specified as required "
                "tags by the plot function: %s",
                ", ".join(required_tags),
            )
            compute_only = required_tags

        # Make sure the compute_only argument contains all the required tags
        elif compute_only is not None and required_tags is not None:
            missing_tags = [t for t in required_tags if t not in compute_only]

            if missing_tags:
                raise ValueError(
                    "Plot function {} required tags that were "
                    "not set to be computed by the DAG: {}. Make "
                    "sure to set the `compute_only` argument "
                    "such that results for all required tags "
                    "({}) will actually be computed.\n"
                    "Available tags:  {}\n"
                    "compute_only:    {}"
                    "".format(
                        _plot_func,
                        ", ".join(missing_tags),
                        ", ".join(required_tags),
                        ", ".join(dag.tags),
                        ", ".join(compute_only),
                    )
                )

        # Now, compute, using the parent method
        return super()._compute_dag(
            dag, compute_only=compute_only, **compute_kwargs
        )

    def _combine_dag_results_and_plot_cfg(
        self,
        *,
        dag: TransformationDAG,
        dag_results: dict,
        dag_params: dict,
        plot_kwargs: dict,
    ) -> dict:
        """Returns a dict of plot configuration and ``data``, where all the
        DAG results are stored in.
        In case where the DAG results are to be unpacked, the DAG results will
        be made available as separate keyword arguments instead of as the
        single ``data`` keyword argument.

        Furthermore, if the plot function specified in its attributes that the
        DAG object is to be passed along, this is the place where it is
        included or excluded from the arguments.

        .. note::

            This behaviour is different than in the parent class, where the
            DAG results are passed on as ``dag_results``.

        """
        if dag_params["unpack_dag_results"]:
            # Unpack the results such that they can be specified in the plot
            # function signature
            try:
                cfg = dict(**dag_results, **plot_kwargs)

            except TypeError as err:
                raise TypeError(
                    "Failed unpacking DAG results! There were arguments of "
                    "the same names as some DAG tags given in the plot "
                    "configuration. Make sure they have unique names or "
                    "disable unpacking of the DAG results.\n"
                    "  Keys in DAG results: {}\n"
                    "  Keys in plot config: {}\n"
                    "".format(
                        ", ".join(dag_results.keys()),
                        ", ".join(plot_kwargs.keys()),
                    )
                ) from err

        else:
            # Make the DAG results available as `data` kwarg
            cfg = dict(data=dag_results, **plot_kwargs)

        # Add the `dag` kwarg, if configured to do so.
        if dag_params["pass_dag_object_along"]:
            cfg["dag"] = dag

        return cfg

    # .........................................................................
    # Helpers: Style and Animation

    def _prepare_style_context(
        self,
        *,
        base_style: Union[str, List[str]] = None,
        rc_file: str = None,
        ignore_defaults: bool = False,
        **update_rc_params,
    ) -> dict:
        """Builds a dictionary with rcparams for use in a matplotlib rc context

        Args:
            base_style (Union[str, List[str]], optional): The matplotlib
                style to use as a basis for the generated rc parameters dict.
            rc_file (str, optional): path to a YAML file containing rc
                parameters. These are used to update those of the base styles.
            ignore_defaults (bool, optional): Whether to ignore the rc
                parameters that were given to the __init__ method
            **update_rc_params: All further parameters update those that are
                already provided by base_style and/or rc_file arguments.

        Returns:
            dict: The rc parameters dictionary, a valid dict to enter a
                matplotlib style context with

        Raises:
            ValueError: On invalid arguments
        """
        # Determine what to base this
        if self._default_rc_params and not ignore_defaults:
            log.debug("Composing RC parameters based on defaults ...")
            rc_dict = self._default_rc_params

        else:
            log.debug("Composing RC parameters ...")
            rc_dict = dict()

        # Make sure base_style is a list of strings
        if not base_style:
            base_style = []

        elif isinstance(base_style, str):
            base_style = [base_style]

        elif not isinstance(base_style, (list, tuple)):
            raise TypeError(
                "Argument `base_style` need be None, a string, "
                f"or a list of strings, was of type {type(base_style)} with "
                f"value '{base_style}'!"
            )

        # Now, base_style definitely is an iterable.
        # Use it to initially populate the RC dict
        if base_style:
            log.debug("Using base styles: %s", ", ".join(base_style))

            # Iterate over it and populate the rc_dict
            for style_name in base_style:
                # If the base_style key is given, load a dictionary with the
                # corresponding rc_params
                if style_name not in plt.style.available:
                    _available = ", ".join(plt.style.available)
                    raise ValueError(
                        f"Style '{style_name}' is not a valid matplotlib "
                        f"style. Available styles: {_available}"
                    )

                rc_dict = recursive_update(
                    rc_dict, plt.style.library[style_name]
                )

        # If a `rc_file` is specifed update the `rc_dict`
        if rc_file:
            path_to_rc = os.path.expanduser(rc_file)

            if not os.path.isabs(path_to_rc):
                raise ValueError(
                    "Argument `rc_file` needs to be an absolute "
                    f"path, was not! Got: {path_to_rc}"
                )

            elif not os.path.exists(path_to_rc):
                raise ValueError(
                    f"No file was found at path {path_to_rc} specified by "
                    "argument `rc_file`!"
                )

            log.debug("Loading RC parameters from file %s ...", path_to_rc)
            rc_dict = recursive_update(rc_dict, load_yml(path_to_rc))

        # If any other rc_params are specified, update the `rc_dict` with them
        if update_rc_params:
            log.debug("Recursively updating RC parameters...")
            rc_dict = recursive_update(rc_dict, update_rc_params)

        return rc_dict

    def _build_style_context(self, **rc_params):
        """Constructs the matplotlib style context manager, if parameters were
        given, otherwise returns the DoNothingContext
        """
        if rc_params:
            log.remark(
                "Using custom style context with %d entries ...",
                len(rc_params),
            )
            return plt.rc_context(rc=rc_params)
        return DoNothingContext()

    def _perform_animation(
        self,
        *,
        hlpr: PlotHelper,
        style_context,
        plot_func: Callable,
        plot_args: tuple,
        plot_kwargs: dict,
        writer: str,
        writer_kwargs: dict = None,
        animation_update_kwargs: dict = None,
    ):
        """Prepares the Writer and checks for valid animation config.

        Args:
            hlpr (PlotHelper): The plot helper
            style_context: The style context to enter before starting animation
            plot_func (Callable): plotting function which is to be animated
            plot_args (tuple): positional arguments to ``plot_func``
            plot_kwargs (dict): keyword arguments to ``plot_func``
            writer (str): name of movie writer with which the frames are saved
            writer_kwargs (dict, optional): A dict of writer parameters. These
                are associated with the chosen writer via the top level key
                in ``writer_kwargs``. Each dictionary container has three
                further keys queried, all optional:

                    init:
                        passed to ``Writer.__init__`` method
                    saving:
                        passed to ``Writer.saving`` method
                    grab_frame:
                        passed to ``Writer.grab_frame`` method

            animation_update_kwargs (dict, optional): Passed to the animation
                update generator call.

        Raises:
            ValueError: if the animation is not supported by the ``plot_func``
                or if the writer is not available
        """
        # Check that the plot function actually supports animation
        if not getattr(plot_func, "supports_animation", False):
            raise ValueError(
                f"Plotting function '{plot_func.__name__}' was not marked as "
                "supporting an animation! To do so, add the "
                "`supports_animation` flag to the plot function "
                "decorator."
            )

        # Get the kwargs for __init__, saving, and grab_frame of the writer
        writer_name = writer
        writer_cfg = (
            writer_kwargs.get(writer_name, {}) if writer_kwargs else {}
        )

        # Need to extract `dpi`, because matplotlib interface wants it as
        # positional argument. Damn you, matplotlib.
        dpi = writer_cfg.get("saving", {}).pop("dpi", 96)

        # Retrieve the writer: Either from matplotlib or dantro's FileWriter
        if mpl.animation.writers.is_available(writer_name):
            wCls = mpl.animation.writers[writer_name]
            writer = wCls(**writer_cfg.get("init", {}))

        else:
            _available = ", ".join(mpl.animation.writers.list())
            raise ValueError(
                f"The writer '{writer_name}' is not available on your "
                f"system! Available writers: {_available}"
            )

        # Now got the writer.

        # Can enter the style context and perform animation now.
        # In order to not aggregate additional figures during this process,
        # also enter an additional context manager, which prevents figures
        # leaking from the plot function or the animation generator.
        leak_prev = figure_leak_prevention(close_current_fig_on_raise=True)

        log.debug(
            "Performing animation of plot function '%s' using writer %s ...",
            plot_func.__name__,
            writer_name,
        )

        with style_context, leak_prev:
            hlpr.setup_figure()

            # Call the plot function
            plot_func(*plot_args, **plot_kwargs)
            # NOTE This plot is NOT saved as the first frame in order to allow
            #      the animation update generator be a more general method.

            # If helpers are _not_ called later, call them now and be done.
            # While they would not need to be kept enabled, doing so causes no
            # harm and is the more expected behaviour.
            if not hlpr.invoke_before_grab:
                hlpr.invoke_enabled(axes="all", mark_disabled_after_use=False)

            # Enter context manager of movie writer
            with writer.saving(
                hlpr.fig, hlpr.out_path, dpi, **writer_cfg.get("saving", {})
            ):

                # Create the iterator for the animation
                log.debug("Invoking animation update generator ...")
                anim_it = hlpr.animation_update(
                    **(
                        animation_update_kwargs
                        if animation_update_kwargs
                        else {}
                    )
                )

                # Create generator and perform the iteration. The return value
                # of the generator currently is ignored.
                log.debug("Iterating animation update generator ...")
                for frame_no, _ in enumerate(anim_it):
                    # Update the figure used in the writer
                    # This is required for cases in which each frame is given
                    # by a new figure.
                    if writer.fig is not hlpr.fig:
                        writer.fig = hlpr.fig

                    # If required, invoke all enabled helpers before grabbing
                    if hlpr.invoke_before_grab:
                        hlpr.invoke_enabled(
                            axes="all", mark_disabled_after_use=False
                        )

                    # The anim_it invocation has already created the new frame.
                    # Grab it; the writer takes care of saving it
                    writer.grab_frame(**writer_cfg.get("grab_frame", {}))
                    log.debug("Grabbed frame %d.", frame_no)

            # Exited 'saving' context
            # Make sure the figure is closed
            hlpr.close_figure()

        # Exited externally given style context and figure_leak_prevention.
        # Done now.
        log.debug("Animation finished after %s frames.", frame_no + 1)

    # .........................................................................
    # Helpers: PlotManager's auto-detection feature

    def _declared_plot_func_by_attrs(
        self, pf: Callable, creator_name: str
    ) -> bool:
        """Checks whether the given function has attributes set that declare
        it as a plotting function that is to be used with this creator.

        Args:
            pf (Callable): The plot function to check attributes of
            creator_name (str): The name under which this creator type is
                registered to the PlotManager.

        Returns:
            bool: Whether the plot function attributes declare the given plot
                function as suitable for working with this specific creator.
        """
        if hasattr(pf, "creator_type") and pf.creator_type is not None:
            if isinstance(self, pf.creator_type):
                log.debug(
                    "The desired type of the plot function, %s, is the "
                    "same or a parent type of this %s.",
                    pf.creator_type,
                    self.logstr,
                )
                return True
        else:
            log.debug(
                "The plot function's specified creator type (%s) does "
                "not match %s.",
                getattr(pf, "creator_type", None),
                self.logstr,
            )

        if hasattr(pf, "creator_name") and pf.creator_name == creator_name:
            log.debug(
                "The plot function's desired creator name '%s' "
                "matches the name under which %s is known to the "
                "PlotManager.",
                pf.creator_name,
                self.classname,
            )
            return True
        else:
            log.debug(
                "The plot function's specified creator name (%s) does "
                "not match the specified creator name '%s' of %s.",
                getattr(pf, "creator_name", None),
                creator_name,
                self.logstr,
            )

        log.debug(
            "Checked plot function attributes, but neither the type "
            "nor the creator name were specified or matched this "
            "creator."
        )
        return False


# -----------------------------------------------------------------------------


class is_plot_func:
    """This is a decorator class declaring the decorated function as a
    plotting function to use with ExternalPlotCreator-derived plot creators
    """

    def __init__(
        self,
        *,
        creator_type: type = None,
        creator_name: str = None,
        use_helper: bool = True,
        helper_defaults: Union[dict, str] = None,
        use_dag: bool = None,
        required_dag_tags: Sequence[str] = None,
        compute_only_required_dag_tags: bool = True,
        pass_dag_object_along: bool = False,
        unpack_dag_results: bool = False,
        supports_animation=False,
        add_attributes: dict = None,
    ):
        """Initialize the decorator. Note that the function to be decorated is
        not passed to this method.

        Args:
            creator_type (type, optional): The type of plot creator to use
            creator_name (str, optional): The name of the plot creator to use
            use_helper (bool, optional): Whether to use a PlotHelper
            helper_defaults (Union[dict, str], optional): Default
                configurations for helpers; these are automatically considered
                to be enabled. If string-like, will assume this is an absolute
                path to a YAML file and will load the dict-like configuration
                from there.
            use_dag (bool, optional): Whether to use the data transformation
                framework.
            required_dag_tags (Sequence[str], optional): The DAG tags that are
                required by the plot function.
            compute_only_required_dag_tags (bool, optional): Whether to compute
                only those DAG tags that are specified as required by the plot
                function. This is ignored if no required DAG tags were given
                and can be overwritten by the ``compute_only`` argument.
            pass_dag_object_along (bool, optional): Whether to pass on the DAG
                object to the plot function
            unpack_dag_results (bool, optional): Whether to unpack the results
                of the DAG computation directly into the plot function instead
                of passing it as a dictionary.
            supports_animation (bool, optional): Whether the plot function
                supports animation.
            add_attributes (dict, optional): Additional attributes to add to
                the plot function.

        Raises:
            ValueError: If `helper_defaults` was a string but not an absolute
                path.
        """
        if isinstance(helper_defaults, str):
            # Interpret as path to yaml file
            fpath = os.path.expanduser(helper_defaults)

            # Should be absolute
            if not os.path.isabs(fpath):
                raise ValueError(
                    "`helper_defaults` string argument was a "
                    "relative path: {}, but needs to be either a "
                    "dict or an absolute path (~ allowed)."
                    "".format(fpath)
                )

            log.debug("Loading helper defaults from file %s ...", fpath)
            helper_defaults = load_yml(fpath)

        # Gather those attributes that are to be set as function attributes
        self.pf_attrs = dict(
            creator_type=creator_type,
            creator_name=creator_name,
            use_helper=use_helper,
            helper_defaults=helper_defaults,
            use_dag=use_dag,
            required_dag_tags=required_dag_tags,
            compute_only_required_dag_tags=compute_only_required_dag_tags,
            pass_dag_object_along=pass_dag_object_along,
            supports_animation=supports_animation,
            **(add_attributes if add_attributes else {}),
        )

    def __call__(self, func: Callable):
        """If there are decorator arguments, __call__() is only called
        once, as part of the decoration process and expects as only argument
        the function to be decorated.
        """
        # Do not actually wrap the function call, but add attributes to it
        for k, v in self.pf_attrs.items():
            setattr(func, k, v)

        # Return the function, now with attributes set
        return func



"""Private low-level helper classes and functions for the DAG framework

NOTE This is imported by dantro.tools to register classes with YAML.
"""
import logging
from typing import Any, Tuple, Union

from paramspace.tools import recursive_collect, recursive_replace

# Local constants
log = logging.getLogger(__name__)


# -----------------------------------------------------------------------------

# Contains hooks that are invoked when a certain operation is parsed.
#
# The values should be callables that receive ``operation, *args, **kwargs``
# and return a 3-tuple of the manipulated ``operation, args, kwargs``.
# The return values will be those that the Transformation object is created
# from.
#
# Example of defining a hook and registering it:
#
# .. code-block:: python
#
#     def _op_hook_my_operation(operation, *args, **kwargs
#                               ) -> Tuple[str, list, dict]:
#         """An operation hook for my_operation"""
#         # ... do stuff ...
#         return operation, args, kwargs
#
#     DAG_PARSER_OPERATION_HOOKS['my_operation'] = _op_hook_my_operation
#
DAG_PARSER_OPERATION_HOOKS = dict()


# -----------------------------------------------------------------------------


class Placeholder:
    """A generic placeholder class for use in the DAG framework.

    Objects of this class or derived classes are yaml-representable and thus
    hashable after a parent object created a YAML representation. In addition,
    the __hash__ method can be used to generate a hash from the string
    representation.
    """

    def __init__(self, data: Any):
        """Initialize a Placeholder by storing its payload"""
        self._data = data

    def __eq__(self, other) -> bool:
        """Only objects with exactly the same type and data are regarded as
        equal; specifically, this makes instances of subclasses always unequal
        to instances of this base class.
        """
        if type(other) == type(self):
            return self._data == other._data
        return False

    def __repr__(self) -> str:
        return "<{} {}>".format(type(self).__name__, repr(self._data))

    def __hash__(self) -> int:
        return hash(repr(self))

    # YAML representation . . . . . . . . . . . . . . . . . . . . . . . . . . .
    yaml_tag = "!dag_placeholder"

    @classmethod
    def from_yaml(cls, constructor, node):
        """Construct a Placeholder from a scalar YAML node"""
        return cls(constructor.construct_scalar(node))

    @classmethod
    def to_yaml(cls, representer, node):
        """Create a YAML representation of a Placeholder, carrying only the
        _data attribute over...

        As YAML expects scalar data to be str-like, a type cast is done. The
        subclasses that rely on certain argument types should take care that
        their __init__ method can parse arguments that are str-like.
        """
        return representer.represent_scalar(cls.yaml_tag, str(node._data))


class ResultPlaceholder(Placeholder):
    """A placeholder class for a data transformation result"""

    yaml_tag = "!dag_result"

    @property
    def result_name(self) -> str:
        """The name of the transformation result this is a placeholder for"""
        return self._data


def resolve_placeholders(
    d: dict,
    *,
    dag: "TransformationDAG",
    Cls: type = ResultPlaceholder,
    **compute_kwargs,
) -> dict:
    """Recursively replaces placeholder objects throughout the given dict.

    Computes :py:class:`~dantro.dag.TransformationDAG` results and replaces
    the placeholder objects with entries from the results dict, thereby
    making it possible to compute configuration values using results of the
    `data transformation framework <dag_framework>`, for example as done in
    the plotting framework; see :ref:`dag_result_placeholder`.

    .. warning::

        While this function has a return value, it resolves the placeholders
        in-place, such that the given ``d`` will be mutated even if the return
        value is ignored on the calling site.

    Args:
        d (dict): The object to replace placeholders in. Will recursively walk
            through all dict- and list-like objects to find placeholders.
        dag (TransformationDAG): The data transformation tree to resolve the
            placeholders' results from.
        Cls (type, optional): The expected type of the placeholders.
        **compute_kwargs: Passed on to
            :py:meth:`~dantro.dag.TransformationDAG.compute`.
    """
    # First, collect the placeholders
    is_placeholder = lambda obj: isinstance(obj, Cls)
    phs = recursive_collect(d, select_func=is_placeholder)

    # If there weren't any, don't have anything to do
    if not phs:
        log.remark("No placeholders found to resolve.")
        return d

    # Otherwise, get ready for computing results and resolving placeholders
    to_compute = {ph.result_name for ph in phs}

    log.info(
        "Resolving %d placeholder%s ...",
        len(phs),
        "s" if len(phs) != 1 else "",
    )

    try:
        results = dag.compute(compute_only=list(to_compute), **compute_kwargs)

    except ValueError as exc:
        _ph_names = ", ".join(to_compute)
        raise ValueError(
            "Placeholder resolution failed for one or more of the specified "
            f"placeholder names ({_ph_names})!\n{exc}"
        ) from exc

    except RuntimeError as exc:
        _ph_names = ", ".join(to_compute)
        raise RuntimeError(
            "Placeholder resolution failed due to an error during "
            "computation of the transformation result for any of "
            f"the placeholder tags ({_ph_names})!\n{exc}"
        ) from exc

    d = recursive_replace(
        d,
        select_func=is_placeholder,
        replace_func=lambda p: results[p.result_name],
    )
    log.remark("Finished resolving placeholders.")
    return d


# -----------------------------------------------------------------------------
# Used in meta-operations


class PositionalArgument(Placeholder):
    """A PositionalArgument is a placeholder that holds as payload a positional
    argument's position. This is used, e.g., for meta-operation specification.
    """

    yaml_tag = "!arg"

    def __init__(self, pos: int):
        """Initialize from an integer, also accepting int-convertibles"""
        if not isinstance(pos, int):
            # Need an integer conversion to accept YAML string dumps
            try:
                pos = int(pos)
            except:
                raise TypeError(
                    "PositionalArgument requires an "
                    f"int-convertible argument, got {type(pos)} "
                    f"with value {repr(pos)}!"
                )

        if pos < 0:
            raise ValueError(
                "PositionalArgument requires a non-negative "
                f"position, got {pos}!"
            )

        self._data = pos

    @property
    def position(self) -> int:
        return self._data


class KeywordArgument(Placeholder):
    """A KeywordArgument is a placeholder that holds as payload the name of a
    keyword argument. This is used, e.g., for meta-operation specification.
    """

    yaml_tag = "!kwarg"

    def __init__(self, name: str):
        """Initialize by storing the keyword argument name"""
        if not isinstance(name, str):
            raise TypeError(
                "KeywordArgument requires a string "
                f"as argument name, got {type(name)}!"
            )

        self._data = name

    @property
    def name(self) -> int:
        return self._data


# -----------------------------------------------------------------------------


class DAGReference(Placeholder):
    """The DAGReference class is the base class of all DAG reference objects.
    It extends the generic Placeholder class with the ability to resolve
    references within a :py:class:`~dantro.dag.TransformationDAG`.
    """

    yaml_tag = "!dag_ref"

    def __init__(self, ref: str):
        """Initialize a DAGReference object from a hash."""
        if not isinstance(ref, str):
            raise TypeError(
                "DAGReference requires a string-like argument, "
                f"got {type(ref)}!"
            )

        self._data = ref

    @property
    def ref(self) -> str:
        """The associated reference of this object"""
        return self._data

    def _resolve_ref(self, *, dag: "TransformationDAG") -> str:
        """Return the hash reference; for the base class, the data is already
        the hash reference, so no DAG is needed. Derived classes _might_ need
        the DAG to resolve their reference hash.
        """
        return self._data

    def convert_to_ref(self, *, dag: "TransformationDAG") -> "DAGReference":
        """Create a new object that is a hash ref to the same object this
        tag refers to."""
        return DAGReference(self._resolve_ref(dag=dag))

    def resolve_object(self, *, dag: "TransformationDAG") -> Any:
        """Resolve the object by looking up the reference in the DAG's object
        database.
        """
        return dag.objects[self._resolve_ref(dag=dag)]


# .............................................................................


class DAGTag(DAGReference):
    """A DAGTag object stores a name of a tag, which serves as a named
    reference to some object in the DAG.
    """

    yaml_tag = "!dag_tag"

    def __init__(self, name: str):
        """Initialize a DAGTag object, storing the specified field name"""
        # Prohibit certain names that would collide with DAGMetaOperationTag
        if DAGMetaOperationTag.SPLIT_STR in name:
            raise ValueError(
                "DAGTag names cannot include the "
                f"'{DAGMetaOperationTag.SPLIT_STR}' substring! "
                f"Adjust the name of tag '{name}' accordingly."
            )

        self._data = name

    @property
    def name(self) -> str:
        """The name of the tag within the DAG that this object references"""
        return self._data

    def _resolve_ref(self, *, dag: "TransformationDAG") -> str:
        """Return the hash reference by looking up the tag in the DAG"""
        return dag.tags[self.name]


class DAGMetaOperationTag(DAGTag):
    """A DAGMetaOperationTag stores a name of a tag, just as DAGTag, but can
    only be used inside a meta-operation. When resolving this tag's reference,
    the target is looked up from the stack of the TransformationDAG.
    """

    yaml_tag = "!mop_tag"
    SPLIT_STR = "::"

    def __init__(self, name: str):
        """Initialize the DAGMetaOperationTag object.

        The ``name`` needs to be of the ``<meta-operation name>::<tag name>``
        pattern and thereby include information on the name of the
        meta-operation this tag is used in.
        """
        # Check if valid
        try:
            mop, tag = name.split(self.SPLIT_STR)
        except Exception as exc:
            raise ValueError(
                f"Invalid name '{name}' for DAGMetaOperationTag! "
                f"The '{self.SPLIT_STR}' substring "
                "is missing or is used more than once!"
            ) from exc

        self._data = name

    def _resolve_ref(self, *, dag: "TransformationDAG") -> str:
        """Return the hash reference by looking it up in the reference stacks
        of the specified TransformationDAG. The last entry always refers to the
        currently active meta-operation.
        """
        return dag.ref_stacks[self.name][-1]

    @classmethod
    def make_name(cls, meta_operation: str, *, tag: str) -> str:
        """Given a meta-operation name and a tag name, generates the name of
        this meta-operation tag.
        """
        return f"{meta_operation}{cls.SPLIT_STR}{tag}"

    @classmethod
    def from_names(
        cls, meta_operation: str, *, tag: str
    ) -> "DAGMetaOperationTag":
        """Generates a DAGMetaOperationTag using the names of a meta-operation
        and the name of a tag.
        """
        return cls(cls.make_name(meta_operation, tag=tag))


# .............................................................................


class DAGNode(DAGReference):
    """A DAGNode is a reference by the index within the DAG's node list."""

    yaml_tag = "!dag_node"

    def __init__(self, idx: int):
        """Initialize a DAGNode object with a node index.

        Args:
            idx (int): The idx value to set this reference to. Can also be a
                negative value, in which case the node list is traversed from
                the back.

        Raises:
            TypeError: On invalid type (not int-convertible)
        """
        if not isinstance(idx, int):
            # Need an integer conversion to accept YAML string dumps
            try:
                idx = int(idx)
            except:
                raise TypeError(
                    "DAGNode requires an int-convertible "
                    f"argument, got {type(idx)} with "
                    f"value {repr(idx)}!"
                )

        self._data = idx

    @property
    def idx(self) -> int:
        """The idx to the referenced node within the DAG's node list"""
        return self._data

    def _resolve_ref(self, *, dag: "TransformationDAG") -> str:
        """Return the hash reference by looking up the node index in the DAG"""
        return dag.nodes[self.idx]


# -----------------------------------------------------------------------------


class DAGObjects:
    """An objects database for the DAG framework.

    It uses a flat dict containing (hash, object ref) pairs. The interface is
    slightly restricted compared to a regular dict; especially, item deletion
    is not made available.

    Objects are added to the database via the ``add_object`` method. They need
    to have a ``hashstr`` property, which returns a hash string
    deterministically representing the object; note that this is not
    equivalent to the Python builtin hash() function which invokes the __hash__
    magic method.
    """

    def __init__(self):
        """Initialize an empty objects database"""
        self._d = dict()

    def __str__(self) -> str:
        """A human-readable string representation of the object database"""
        return "<DAGObjects database with {:d} entr{}>".format(
            len(self), "ies" if len(self) != 1 else "y"
        )

    def add_object(self, obj, *, custom_hash: str = None) -> str:
        """Add an object to the object database, storing it under its hash.

        Note that the object cannot be just any object that is hashable but it
        needs to return a string-based hash via the ``hashstr`` property. This
        is a dantro DAG framework-internal interface.

        Also note that the object will NOT be added if an object with the same
        hash is already present. The object itself is of no importance, only
        the returned hash is.

        Args:
            obj: Some object that has the ``hashstr`` property, i.e. is
                hashable as required by the DAG interface
            custom_hash (str, optional): A custom hash to use instead of the
                hash extracted from ``obj``. Can only be given when ``obj``
                does *not* have a ``hashstr`` property.

        Returns:
            str: The hash string of the given object. If a custom hash string
                was given, it is also the return value

        Raises:
            TypeError: When attempting to pass ``custom_hash`` while ``obj``
                *has* a ``hashstr`` property
            ValueError: If the given ``custom_hash`` already exists.
        """
        if custom_hash is not None:
            if hasattr(obj, "hashstr"):
                raise TypeError(
                    "Cannot use a custom hash for objects that provide their "
                    f"own `hashstr` property! Got object of type {type(obj)} "
                    f"and custom hash '{custom_hash}'."
                )

            elif custom_hash in self:
                raise ValueError(
                    f"The provided custom hash '{custom_hash}' for object of "
                    f"type {type(obj)} already exists! Refusing to add it. "
                    "Was the object already added? If not, choose a different "
                    "custom hash."
                )

            key = custom_hash

        else:
            # Use the DAG framework's internal hash method
            key = obj.hashstr

        # Only add the new object, if the hash does not exist yet.
        if key not in self:
            self._d[key] = obj
        return key

    def __getitem__(self, key: str) -> object:
        """Return the object associated with the given hash"""
        return self._d[key]

    def __len__(self) -> int:
        """Returns the number of objects in the objects database"""
        return len(self._d)

    def __contains__(self, key: str) -> bool:
        """Whether the given hash refers to an object in this database"""
        return key in self._d

    def keys(self):
        return self._d.keys()

    def values(self):
        return self._d.values()

    def items(self):
        return self._d.items()


# -----------------------------------------------------------------------------


def parse_dag_minimal_syntax(params: Union[str, dict]) -> dict:
    """Parses the minimal syntax parameters, effectively translating a string-
    like argument to a dict with the string specified as the ``operation`` key.
    """
    if isinstance(params, dict):
        return params

    elif isinstance(params, str):
        return dict(operation=params, with_previous_result=True)

    # else:
    raise TypeError(
        "Expected either dict or string for minimal syntax, got "
        f"{type(params)} with value: {params}"
    )


def parse_dag_syntax(
    *,
    operation: str = None,
    args: list = None,
    kwargs: dict = None,
    tag: str = None,
    with_previous_result: bool = False,
    salt: int = None,
    file_cache: dict = None,
    ignore_hooks: bool = False,
    **ops,
) -> dict:
    """Given the parameters of a transform operation, possibly in a shorthand
    notation, returns a dict with normalized content by expanding the
    shorthand notation.

    Keys that will be available in the resulting dict:
        ``operation``, ``args``, ``kwargs``, ``tag``.

    Args:
        operation (str, optional): Which operation to carry out; can only be
            specified if there is no ``ops`` argument.
        args (list, optional): Positional arguments for the operation; can
            only be specified if there is no ``ops`` argument.
        kwargs (dict, optional): Keyword arguments for the operation; can only
            be specified if there is no ``ops`` argument.
        tag (str, optional): The tag to attach to this transformation
        with_previous_result (bool, optional): Whether the result of the
            previous transformation is to be used as first positional argument
            of this transformation.
        salt (int, optional): A salt to the Transformation object, thereby
            changing its hash.
        file_cache (dict, optional): File cache parameters
        ignore_hooks (bool, optional): If True, there will be no lookup in the
            operation hooks.
        **ops: The operation that is to be carried out. May contain one and
            only one operation.

    Returns:
        dict: The normalized dict of transform parameters.

    Raises:
        ValueError: For invalid notation, e.g. unambiguous specification of
            arguments or the operation.
    """

    def _raise_error(mode: type, *, operation: str, op_params):
        if mode is dict:
            kind, arg_name = "keyword", "kwargs"
        else:
            kind, arg_name = "positional", "args"

        raise ValueError(
            f"Got superfluous `{arg_name}` argument!"
            f"When specifying {kind} arguments via the shorthand notation "
            f"('{operation}: {op_params}'), there can be no additional "
            f"`{arg_name}` argument specified! Remove that argument or merge "
            "its content with the arguments specified via the shorthand."
        )

    # Distinguish between explicit and shorthand mode
    if operation and not ops:
        # Explicit parametrization
        args = args if args else []
        kwargs = kwargs if kwargs else {}

    elif ops and not operation:
        # Shorthand parametrization
        # Make sure there are no stray argument
        if len(ops) > 1:
            raise ValueError(
                "For shorthand notation, there can only be a "
                "single operation specified, but got multiple "
                f"operations: {ops}"
            )

        # Extract operation name and parameters
        operation, op_params = list(ops.items())[0]

        # Depending on type, regard parameters as args or kwargs. If the
        # argument is not a container, assume it's a single positional
        # argument. The arguments that are not specified by op_params will be
        # set from the existing
        if isinstance(op_params, dict):
            if kwargs:
                _raise_error(dict, operation=operation, op_params=op_params)
            args = args if args else []
            kwargs = op_params

        elif isinstance(op_params, (list, tuple)):
            if args:
                _raise_error(list, operation=operation, op_params=op_params)
            args = list(op_params)
            kwargs = kwargs if kwargs else {}

        elif op_params is not None:
            if args:
                _raise_error(list, operation=operation, op_params=op_params)
            args = [op_params]
            kwargs = kwargs if kwargs else {}

        else:
            args = args if args else []
            kwargs = kwargs if kwargs else {}

    elif not operation and not ops:
        raise ValueError(
            "Missing operation specification. Either use the "
            "`operation` key to specify one or use shorthand "
            "notation by using the name of the operation as a "
            "key and adding the arguments to it as values."
        )

    else:
        raise ValueError(
            "Got two specifications of operations, one via the "
            f"`operation` argument ('{operation}'), another via "
            f"the shorthand notation ({ops}). Remove one!"
        )

    # Have variables operation, args, and kwargs set now.

    # If the result is to be carried on, the first _positional_
    # argument is set to be a reference to the previous node
    if with_previous_result:
        args.insert(0, DAGNode(-1))

    # Invoke operation-specific hooks
    if not ignore_hooks and operation in DAG_PARSER_OPERATION_HOOKS:
        hook = DAG_PARSER_OPERATION_HOOKS[operation]
        log.remark("Invoking parser hook for operation '%s' ...", operation)
        try:
            operation, args, kwargs = hook(operation, *args, **kwargs)
        except Exception as exc:
            log.warning(
                "Failed applying operation-specific hook for '%s'! "
                "Got %s: %s.\nEither correct the error or disable "
                "the hook for this operation by setting the "
                "``ignore_hooks`` flag. Otherwise, this operation "
                "might fail during computation.",
                operation,
                exc.__class__.__name__,
                exc,
            )

    # Done. Construct the dict.
    # Mandatory parameters
    d = dict(operation=operation, args=args, kwargs=kwargs, tag=tag)

    # Add optional parameters only if they were specified
    if salt is not None:
        d["salt"] = salt

    if file_cache is not None:
        d["file_cache"] = file_cache

    return d


# -----------------------------------------------------------------------------
# Operation Hooks
# NOTE:
#   - Names should follow ``op_hook_<operation-name>``
#   - A documentation entry should be added in doc/data_io/dag_op_hooks.rst


def op_hook_expression(operation, *args, **kwargs) -> Tuple[str, list, dict]:
    """An operation hook for the ``expression`` operation, attempting to
    auto-detect which symbols are specified in the given expression.
    From those, ``DAGTag`` objects are created, making it more convenient to
    specify an expression that is based on other DAG tags.

    The detected symbols are added to the ``kwargs.symbols``, if no symbol of
    the same name is already explicitly defined there.

    This hook accepts as positional arguments both the ``(expr,)`` form and
    the ``(prev_node, expr)`` form, making it more robust when the
    ``with_previous_result`` flag was set.

    If the expression contains the ``prev`` or ``previous_result`` symbols,
    the corresponding :py:class:`~dantro._dag_utils.DAGNode` will be added to
    the symbols additionally.

    For more information on operation hooks, see :ref:`dag_op_hooks`.
    """
    from sympy import Symbol
    from sympy.parsing.sympy_parser import parse_expr

    # Extract the expression string
    if len(args) == 1:
        expr = args[0]
    elif len(args) == 2:
        _, expr = args
    else:
        raise TypeError(
            f"Got unexpected positional arguments: {args}; expected either "
            "(expr,) or (prev_node, expr)."
        )

    # Try to extract all symbols from the expression
    all_symbols = parse_expr(expr, evaluate=False).atoms(Symbol)

    # Some symbols might already be given; only add those that were not given.
    # Also, convert the ``prev`` and ``previous_result`` symbols the
    # corresponding DAGNode object
    symbols = kwargs.get("symbols", {})
    for symbol in all_symbols:
        symbol = str(symbol)
        if symbol in symbols:
            log.remark(
                "Symbol '%s' was already specified explicitly! It "
                "will not be replaced.",
                symbol,
            )
            continue

        if symbol in ("prev", "previous_result"):
            symbols[symbol] = DAGNode(-1)
        else:
            symbols[symbol] = DAGTag(symbol)

    # For the case of missing ``symbols`` key, need to write it back to kwargs
    kwargs["symbols"] = symbols

    # For args, return _only_ ``expr``, as expected by the operation
    return operation, (expr,), kwargs


DAG_PARSER_OPERATION_HOOKS["expression"] = op_hook_expression



"""Tools for module importing, e.g. lazy imports."""

import logging
from importlib import import_module as _import_module

log = logging.getLogger(__name__)


def get_from_module(mod, *, name: str):
    """Retrieves an attribute from a module, if necessary traversing along the
    module string.

    Args:
        mod: Module to start looking at
        name (str): The ``.``-separated module string leading to the desired
            object.
    """
    obj = mod

    for attr_name in name.split("."):
        try:
            obj = getattr(obj, attr_name)

        except AttributeError as err:
            raise AttributeError(
                f"Failed to retrieve attribute or attribute sequence '{name}' "
                f"from module '{mod.__name__}'! Intermediate "
                f"{type(obj).__name__} {obj} has no attribute '{attr_name}'!"
            ) from err

    return obj


def import_module_or_object(module: str = None, name: str = None):
    """Imports a module or an object using the specified module string and the
    object name.

    Args:
        module (str, optional): A module string, e.g. numpy.random. If this is
            not given, it will import from the :py:mod`builtins` module. Also,
            relative module strings are resolved from :py:mod:`dantro`.
        name (str, optional): The name of the object to retrieve from the
            chosen module and return. This may also be a dot-separated sequence
            of attribute names which can be used to traverse along attributes.

    Returns:
        The chosen module or object, i.e. the object found at <module>.<name>

    Raises:
        AttributeError: In cases where part of the ``name`` argument could not
            be resolved due to a bad attribute name.
    """
    module = module if module else "builtins"
    mod = _import_module(module, package="dantro") if module else __builtin__

    if not name:
        return mod
    return get_from_module(mod, name=name)


class LazyLoader:
    """Delays import until the module's attributes are accessed.

    This is inspired by the implementation by Dboy Liao:
        https://levelup.gitconnected.com/python-trick-lazy-module-loading-df9b9dc111af

    It extends on it by allowing a ``depth`` until which loading will be lazy.
    """

    def __init__(self, mod_name: str, *, _depth: int = 0):
        """Initialize a placeholder for a module.

        .. warning::

            Values of ``_depth > 0`` may lead to unexpected behaviour of the
            root module, i.e. this object, because attribute calls do not
            yield an actual object. Only use this in scenarios where you are
            in full control over the attribute calls.

            We furthermore suggest to not make the LazyLoader instance publicly
            available in such cases.

        Args:
            mod_name (str): The module name to lazy-load upon attribute call.
            _depth (int, optional): With a depth larger than zero, attribute
                calls are not leading to an import *yet*, but to the creation
                of another LazyLoader instance (with depth reduced by one).
                Note the warning above regarding usage.
        """
        self.mod_name = mod_name
        self._mod = None
        self._depth = _depth
        log.debug("LazyLoader set up for '%s'.", self.mod_name)

    def __getattr__(self, name: str):
        # If depth is zero, do the import
        if self._depth <= 0:
            if self._mod is None:
                self._mod = self.resolve()
            return getattr(self._mod, name)

        # else: create a new lazy loader for the combined module string
        return LazyLoader(f"{self.mod_name}.{name}", _depth=self._depth - 1)

    def resolve(self):
        log.debug("Now lazy-loading '%s' ...", self.mod_name)

        modstr_parts = self.mod_name.split(".")
        return import_module_or_object(
            module=modstr_parts[0], name=".".join(modstr_parts[1:])
        )



"""This module implement mixin classes that provide indexing capabilities"""

import logging
from typing import Union

from ..abc import AbstractDataContainer, AbstractDataGroup
from .base import ItemAccessMixin

# Local constants
log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


class IntegerItemAccessMixin:
    """This mixin allows accessing items via integer keys and also supports
    calling the ``__contains__`` magic method with integer keys. It is meant to
    be used to add features to an AbstractDataGroup-derived class, although
    this is not enforced.

    .. note::

        The ``__setitem__`` method is not covered by this!

    .. note::

        The class using this mixin has to implement index access methods and
        the ``__contains__`` magic method independently from this mixin!
    """

    def _parse_key(self, key: Union[str, int]) -> str:
        """Makes sure a key is a string"""
        if isinstance(key, int):
            return str(key)
        return key

    def __getitem__(self, key: Union[str, int]):
        """Adjusts the parent method to allow integer key item access"""
        return super().__getitem__(self._parse_key(key))
        # NOTE Don't need two-argument super() here because the parents are
        #      deduced when the class is assembled, i.e. when mixins and other
        #      inheritances are all combined.

    def __setitem__(self, key: Union[str, int]):
        """Adjusts the parent method to allow item setting by integer key"""
        return super().__setitem__(self._parse_key(key))

    def __delitem__(self, key: Union[str, int]):
        """Adjusts the parent method to allow item deletion by integer key"""
        return super().__delitem__(self._parse_key(key))

    def __contains__(self, key: Union[str, int]) -> bool:
        """Adjusts the parent method to allow checking for integers"""
        return super().__contains__(self._parse_key(key))


class PaddedIntegerItemAccessMixin(IntegerItemAccessMixin):
    """This mixin allows accessing items via integer keys that map to members
    that have a zero-padded integer name. It can only be used as mixin for
    AbstractDataGroup-derived classes!

    The ``__contains__`` magic method is also supported in this mixin.

    .. note::

        The class using this mixin has to implement index access methods and
        the ``__contains__`` magic method independently from this mixin!
    """

    # The number of digits of the padded string representing the integer
    _PADDED_INT_KEY_WIDTH = None

    # The format string to generate a padded integer; deduced upon first call
    _PADDED_INT_FSTR = None

    # Whether to use strict checking when parsing keys, i.e.: check that the
    # range of keys is valid and an error is thrown when an integer key was
    # given that cannot be represented consistently by a padded string of the
    # determined key width.
    _PADDED_INT_STRICT_CHECKING = True

    # The allowed maximum value of an integer key; checked only in strict mode
    _PADDED_INT_MAX_VAL = None

    # .........................................................................

    @property
    def padded_int_key_width(self) -> Union[int, None]:
        """Returns the width of the zero-padded integer key or None, if it is
        not already specified.
        """
        return self._PADDED_INT_KEY_WIDTH

    @padded_int_key_width.setter
    def padded_int_key_width(self, key_width: int):
        """Method to manually provide an integer key width

        Args:
            key_width (int): The

        Raises:
            ValueError: Description
        """
        if self._PADDED_INT_FSTR:
            raise ValueError(
                "Padded integer key width is already set for {}; "
                "cannot set it again!".format(self.logstr)
            )

        elif key_width <= 0:
            raise ValueError(
                "Argument `key_width` to padded_int_key_width "
                "setter property of {} needs to be positive, was "
                "'{}'!".format(self.logstr, key_width)
            )

        # Deduce the key width by going over all member names
        self._PADDED_INT_KEY_WIDTH = key_width

        # Assemble the format string to something like {:05d}
        self._PADDED_INT_FSTR = "{:0" + str(self._PADDED_INT_KEY_WIDTH) + "d}"

        # Compute the maximum value that is fully representable by the string
        self._PADDED_INT_MAX_VAL = 10 ** self._PADDED_INT_KEY_WIDTH - 1

    # .........................................................................

    def _parse_key(self, key: Union[str, int]) -> str:
        """Parse a potentially integer key to a zero-padded string"""
        # Check if it even is an integer
        if not isinstance(key, int):
            return key

        # Also, cannot work properly if no format string was generated, i.e. no
        # key width was provided. This can be the case if there are no members
        # added to the group using this mixin yet
        if not self._PADDED_INT_FSTR:
            return str(key)

        # Optionally, make an additional check for integer key values
        if self._PADDED_INT_STRICT_CHECKING:
            if key < 0 or key > self._PADDED_INT_MAX_VAL:
                raise IndexError(
                    "Integer index {} out of range [0, {}] for {}!".format(
                        key, self._PADDED_INT_MAX_VAL, self.logstr
                    )
                )

        # Generate the key string from the format string
        return self._PADDED_INT_FSTR.format(key)

    def _check_cont(self, cont: AbstractDataContainer) -> None:
        """This method is invoked when adding a member to a group and makes
        sure the name of the added group is correctly zero-padded.

        Also, upon first call, communicates the zero padded integer key width,
        i.e.: the length of the container name, to the
        PaddedIntegerItemAccessMixin.

        Args:
            cont: The member container to add

        Returns
            None: No return value needed
        """
        if self.padded_int_key_width is None:
            self.padded_int_key_width = len(cont.name)

        if len(cont.name) != self.padded_int_key_width:
            raise ValueError(
                f"All containers that are to be added to {self.logstr} need "
                f"names of the same length ({self.padded_int_key_width}), "
                f"but {cont.logstr} had a name of length {len(cont.name)}!"
            )

        # Still invoke the potentially existing parent method
        super()._check_cont(cont)



"""This module implements the VegaPlotCreator class"""

import logging

from .pcr_base import BasePlotCreator

# Local constants
log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


class VegaPlotCreator(BasePlotCreator):
    """This PlotCreator interfaces with Altair to provide a Vega-Lite interface
    for plot creation.
    """

    pass



"""Takes care of all YAML-related imports and configuration

The ``ruamel.yaml.YAML`` object used here is imported from ``paramspace`` and
specialized such that it can load and dump dantro classes.
"""

import io
import logging
import os
from typing import Any, Union

import ruamel.yaml

# Local constants
log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------

# Import yaml from paramspace and configure it
from paramspace import yaml

yaml.default_flow_style = False

# Register further classes
from ._dag_utils import (
    DAGNode,
    DAGReference,
    DAGTag,
    KeywordArgument,
    Placeholder,
    PositionalArgument,
    ResultPlaceholder,
)

yaml.register_class(Placeholder)
yaml.register_class(ResultPlaceholder)
yaml.register_class(PositionalArgument)
yaml.register_class(KeywordArgument)
yaml.register_class(DAGReference)
yaml.register_class(DAGTag)
yaml.register_class(DAGNode)

# Special constructors ........................................................
# For the case of a reference to the previous node
yaml.constructor.add_constructor("!dag_prev", lambda l, n: DAGNode(-1))


# -----------------------------------------------------------------------------


def load_yml(path: str, *, mode: str = "r") -> Union[dict, Any]:
    """Deserializes a YAML file into an object.

    Uses the dantro-internal ``ruamel.yaml.YAML`` object for loading and thus
    supports all registered constructors.

    Args:
        path (str): The path to the YAML file that should be loaded. A ``~`` in
            the path will be expanded to the current user's directory.
        mode (str, optional): Read mode

    Returns:
        Union[dict, Any]: The result of the data loading. Typically, this will
            be a dict, but depending on the structure of the file, it may also
            be of another type.
    """
    path = os.path.expanduser(path)
    log.debug("Loading YAML file... mode: %s, path:\n  %s", mode, path)

    with open(path, mode) as yaml_file:
        return yaml.load(yaml_file)


def write_yml(d: Union[dict, Any], *, path: str, mode: str = "w"):
    """Serialize an object using YAML and store it in a file.

    Uses the dantro-internal ``ruamel.yaml.YAML`` object for dumping and thus
    supports all registered representers.

    Args:
        d (dict): The object to serialize and write to file
        path (str): The path to write the YAML output to. A ``~`` in the path
            will be expanded to the current user's directory.
        mode (str, optional): Write mode of the file
    """
    path = os.path.expanduser(path)
    log.debug(
        "Dumping %s to YAML file... mode: %s, path:\n  %s",
        type(d).__name__,
        mode,
        path,
    )

    # Make sure the directory is present
    os.makedirs(os.path.dirname(path), exist_ok=True)

    with open(path, mode) as yaml_file:
        # Add the yaml '---' prefix
        yaml_file.write("---\n")

        # Now dump the rest
        yaml.dump(d, stream=yaml_file)

        # Ensure new line
        yaml_file.write("\n")


# -----------------------------------------------------------------------------
# More specific YAML-based functions


def yaml_dumps(
    obj: Any, *, register_classes: tuple = (), **dump_params
) -> str:
    """Serializes the given object using a newly created YAML dumper.

    The aim of this function is to provide YAML dumping that is not dependent
    on any package configuration; all parameters can be passed here.

    In other words, his function does _not_ use the dantro._yaml.yaml object
    for dumping but each time creates a new dumper with fixed settings. This
    reduces the chance of interference from elsewhere. Compared to the time
    needed for serialization in itself, the extra time needed to create the
    new ruamel.yaml.YAML object and register the classes is negligible.

    Args:
        obj (Any): The object to dump
        register_classes (tuple, optional): Additional classes to register
        **dump_params: Dumping parameters

    Returns:
        str: The output of serialization

    Raises:
        ValueError: On failure to serialize the given object
    """
    s = io.StringIO()
    y = ruamel.yaml.YAML()

    # Register classes; then apply dumping parameters via object properties
    for Cls in register_classes:
        y.register_class(Cls)

    for k, v in dump_params.items():
        setattr(y, k, v)

    # Serialize
    try:
        y.dump(obj, stream=s)

    except Exception as err:
        raise ValueError(
            f"Could not serialize the given {type(obj)} object!"
        ) from err

    return s.getvalue()



"""This module implements loaders mixin classes for use with the
:py:class:`~dantro.data_mngr.DataManager`.

All these mixin classes should follow the following pattern:

.. code-block:: python

    class LoadernameLoaderMixin:

        @add_loader(TargetCls=TheTargetContainerClass)
        def _load_loadername(filepath: str, *, TargetCls: type):
            # ...
            return TargetCls(...)

As ensured by the :py:func:`~dantro.data_loaders._tools.add_loader` decorator
(implemented in :py:mod:`dantro.data_loaders._tools` module), each
``_load_loadername`` method gets supplied with the path to a file and the
``TargetCls`` argument, which can be called to create an object of the correct
type and name.

By default, and to decouple the loader from the container, it should be
considered to be a static method; in other words: the first positional argument
should ideally *not* be ``self``!
If ``self`` is required for some reason, set the ``omit_self`` option of the
decorator to ``False``, making it a regular (instead of a static) method.
"""

from .load_hdf5 import Hdf5LoaderMixin
from .load_numpy import NumpyLoaderMixin
from .load_pkl import PickleLoaderMixin
from .load_text import TextLoaderMixin
from .load_xarray import XarrayLoaderMixin
from .load_yaml import YamlLoaderMixin


class AllAvailableLoadersMixin(
    TextLoaderMixin,
    YamlLoaderMixin,
    PickleLoaderMixin,
    Hdf5LoaderMixin,
    XarrayLoaderMixin,
    NumpyLoaderMixin,
):
    """A mixin bundling all data loaders that are available in dantro.

    This is useful for a more convenient import in a downstream
    :py:class:`~dantro.data_mngr.DataManager`.

    See the individual mixins for a more detailed documentation.
    """

    pass


# fmt: off

# A dict of file extensions and preferred loader names for those extensions
LOADER_BY_FILE_EXT = {
    'txt':      'text',
    'log':      'text',

    'yml':      'yml',
    'yaml':     'yaml',

    'pickle':   'pickle',
    'pkl':      'pkl',

    'hdf5':     'hdf5',
    'h5':       'hdf5',

    'nc':       'xr_dataarray',
    'netcdf':   'xr_dataarray',
    'nc_da':    'xr_dataarray',
    'xrdc':     'xr_dataarray',
    'nc_ds':    'xr_dataset',

    'npy':      'numpy_binary'
}



"""Configures the DantroLogger for the whole package"""

import logging
from logging import getLogger

# Define the additional log levels
TRACE = 5
REMARK = 12
NOTE = 18
PROGRESS = 22
CAUTION = 23
HILIGHT = 25
SUCCESS = 35


class DantroLogger(logging.getLoggerClass()):
    """The custom dantro logging class with additional log levels"""

    def __init__(self, name, level=logging.NOTSET):
        super().__init__(name, level)

        logging.addLevelName(TRACE, "TRACE")
        logging.addLevelName(REMARK, "REMARK")
        logging.addLevelName(NOTE, "NOTE")
        logging.addLevelName(PROGRESS, "PROGRESS")
        logging.addLevelName(CAUTION, "CAUTION")
        logging.addLevelName(HILIGHT, "HILIGHT")
        logging.addLevelName(SUCCESS, "SUCCESS")

    def trace(self, msg, *args, **kwargs):
        if self.isEnabledFor(TRACE):
            self._log(TRACE, msg, args, **kwargs)

    def remark(self, msg, *args, **kwargs):
        if self.isEnabledFor(REMARK):
            self._log(REMARK, msg, args, **kwargs)

    def note(self, msg, *args, **kwargs):
        if self.isEnabledFor(NOTE):
            self._log(NOTE, msg, args, **kwargs)

    def progress(self, msg, *args, **kwargs):
        if self.isEnabledFor(PROGRESS):
            self._log(PROGRESS, msg, args, **kwargs)

    def caution(self, msg, *args, **kwargs):
        if self.isEnabledFor(CAUTION):
            self._log(CAUTION, msg, args, **kwargs)

    def hilight(self, msg, *args, **kwargs):
        if self.isEnabledFor(HILIGHT):
            self._log(HILIGHT, msg, args, **kwargs)

    def success(self, msg, *args, **kwargs):
        if self.isEnabledFor(SUCCESS):
            self._log(SUCCESS, msg, args, **kwargs)


# Configure logging, valid for the whole module
logging.setLoggerClass(DantroLogger)
logging.basicConfig(
    format="%(levelname)-8s %(module)-12s %(message)s",
    level=logging.INFO,
)



"""Defines a data loader for Python pickles."""

import dill

from ..containers import ObjectContainer
from ._tools import add_loader

# -----------------------------------------------------------------------------


class PickleLoaderMixin:
    """Supplies a load function for pickled python objects.

    For unpickling, the dill package is used.
    """

    @add_loader(TargetCls=ObjectContainer, omit_self=False)
    def _load_pickle(
        self, filepath: str, *, TargetCls: type, **pkl_kwargs
    ) -> ObjectContainer:
        """Load a pickled object using ``dill.load``.

        Args:
            filepath (str): Where the pickle-dumped file is located
            TargetCls (type): The class constructor
            **pkl_kwargs: Passed on to the load function

        Returns:
            ObjectContainer: The unpickled object, stored in a dantro container
        """
        # Open file in binary mode and unpickle with the given load function
        with open(filepath, mode="rb") as f:
            obj = dill.load(f, **pkl_kwargs)

        # Populate the target container with the object
        return TargetCls(data=obj, attrs=dict(filepath=filepath))

    # Also make available under `pkl`
    _load_pkl = _load_pickle



"""Implements LabelledDataGroup specializations for time series"""

from .labelled import LabelledDataGroup
from .ordered import IndexedDataGroup

# -----------------------------------------------------------------------------


class TimeSeriesGroup(LabelledDataGroup, IndexedDataGroup):
    """A time-series group assumes that each stored member refers to one
    point in time, where the name is to be interpreted as the time coordinate.

    For more information on selection methods, see:

        * :py:meth:`~dantro.groups.labelled.LabelledDataGroup.sel`
        * :py:meth:`~dantro.groups.labelled.LabelledDataGroup.isel`
    """

    # There is only one dimension in a TimeSeriesGroup: time
    LDG_DIMS = ("time",)

    # Names are expected to be the time coordinates of each container
    LDG_EXTRACT_COORDS_FROM = "name"


class HeterogeneousTimeSeriesGroup(TimeSeriesGroup):
    """This extends the :py:class:`~dantro.groups.time_series.TimeSeriesGroup`
    by configuring it such that it retrieves its coordinates not from the name
    of the members contained in it but from _their_ data.

    It still manages only the ``time`` dimension, which is now overlapping with
    the ``time`` dimension in the members of this group. However, the
    py:class:`~dantro.groups.labelled.LabelledDataGroup` can handle this
    overlap and provides a uniform selection interface that allows combining
    this heterogeneously stored data.

    This becomes especially useful in cases where the members of this group
    store data with the following properties:

        * Potentially different coordiantes than the coordinates of other
          members of the group.
        * Containing time information for more than a single time coordinate
        * No guarantee for overlaps between ``time`` dimension or any other
          dimension.

    As such it is suitable to work with data that represents ensembles that
    frequently change not only their size but also their identifying labels.
    Additionally, it supports them not being stored in regular intervals but
    only upon a change in coordinates.
    """

    # Coordinates are extracted from the data directly, inspecting only the
    # group-level dimensions (here: ``time``)
    LDG_EXTRACT_COORDS_FROM = "data"



"""The containers sub-package implements BaseDataContainer specializations"""

from .general import (
    MutableMappingContainer,
    MutableSequenceContainer,
    ObjectContainer,
    PassthroughContainer,
    StringContainer,
)
from .link import LinkContainer
from .numeric import NumpyDataContainer
from .xrdatactr import XrDataContainer



"""This module implements data processing operations for dantro objects"""

import logging
import math
import operator
import re
from difflib import get_close_matches as _get_close_matches
from typing import Any, Callable, Iterable, List, Sequence, Tuple, Union

import numpy as np
import xarray as xr

from .._import_tools import (
    LazyLoader,
    get_from_module,
    import_module_or_object,
)
from ..base import BaseDataContainer, BaseDataGroup
from ..tools import apply_along_axis, make_columns, recursive_getitem
from .coords import extract_coords_from_attrs, extract_dim_names
from .ordereddict import KeyOrderedDict

# Local constants
log = logging.getLogger(__name__)

# Lazy module imports
nx = LazyLoader("networkx")
scipy = LazyLoader("scipy")


# -----------------------------------------------------------------------------
# Operation Implementations
# NOTE Operations should use only logger levels <= REMARK

# fmt: off
# Define boolean operators separately; registered into _OPERATIONS below
BOOLEAN_OPERATORS = {
    '==': operator.eq,  'eq': operator.eq,
    '<':  operator.lt,  'lt': operator.lt,
    '<=': operator.le,  'le': operator.le,
    '>':  operator.gt,  'gt': operator.gt,
    '>=': operator.ge,  'ge': operator.ge,
    '!=': operator.ne,  'ne': operator.ne,
    '^':  operator.xor, 'xor': operator.xor,
    # Expecting an iterable as second argument
    'in':               (lambda x, y: x in y),
    'not in':           (lambda x, y: x not in y),
    # Performing bitwise boolean operations to support numpy logic
    'in interval':      (lambda x, y: x >= y[0] & x <= y[1]),
    'not in interval':  (lambda x, y: x < y[0] | x > y[1]),
} # End of boolean operator definitions
# fmt: on

# .............................................................................


def print_data(data: Any) -> Any:
    """Prints and passes on the data.

    The print operation distinguishes between dantro types (in which case some
    more information is shown) and  non-dantro types.
    """
    # Distinguish between dantro types and others
    if isinstance(data, BaseDataContainer):
        print("{}, with data:\n{}\n".format(data, data.data))

    elif isinstance(data, BaseDataGroup):
        print("{}\n".format(data.tree))

    else:
        print(data)

    return data


# .............................................................................
# symbolic math and expression parsing


def expression(
    expr: str,
    *,
    symbols: dict = None,
    evaluate: bool = True,
    transformations: Tuple[Callable] = None,
    astype: Union[type, str] = float,
):
    """Parses and evaluates a symbolic math expression using SymPy.

    For parsing, uses sympy's ``parse_expr`` function (see documentation of the
    `parsing module <https://docs.sympy.org/latest/modules/parsing.html>`_).
    The ``symbols`` are provided as ``local_dict``; the ``global_dict`` is not
    explicitly set and subsequently uses the sympy default value, containing
    all basic sympy symbols and notations.

    .. note::

        The expression given here is not Python code, but symbolic math.
        You cannot call arbitrary functions, but only those that are imported
        by ``from sympy import *``.

    .. hint::

        When using this expression as part of the :ref:`dag_framework`, it is
        attached to a so-called :ref:`syntax hook <dag_op_hooks_integration>`
        that makes it easier to specify the ``symbols`` parameter.
        See :ref:`here <dag_op_hook_expression>` for more information.

    .. warning::

        While the expression is symbolic math, be aware that smypy by default
        interprets the ``^`` operator as XOR.
        For exponentiation, use the``**`` operator or adjust the
        ``transformations`` argument as specified in the sympy documentation.

    .. warning::

        While the expression is symbolic math, it uses the ``**`` operator for
        exponentiation, unless a custom ``transformations`` argument is given.

        Thus, the ``^`` operator will lead to an XOR operation being performed!

    .. warning::

        The return object of this operation will *only* contain symbolic sympy
        objects if ``astype is None``. Otherwise, the type cast will evaluate
        all symbolic objects to the numerical equivalent specified by the given
        ``astype``.

    Args:
        expr (str): The expression to evaluate
        symbols (dict, optional): The symbols to use
        evaluate (bool, optional): Controls whether sympy evaluates ``expr``.
            This *may* lead to a fully evaluated result, but does not guarantee
            that no sympy objects are contained in the result. For ensuring
            a fully numerical result, see the ``astype`` argument.
        transformations (Tuple[Callable], optional): The ``transformations``
            argument for sympy's ``parse_expr``. By default, the sympy
            standard transformations are performed.
        astype (Union[type, str], optional): If given, performs a cast to this
            data type, fully evaluating all symbolic expressions.
            Default: Python ``float``.

    Raises:
        TypeError: Upon failing ``astype`` cast, e.g. due to free symbols
            remaining in the evaluated expression.
        ValueError: When parsing of ``expr`` failed.

    Returns:
        The result of the evaluated expression.
    """
    from sympy.parsing.sympy_parser import parse_expr as _parse_expr
    from sympy.parsing.sympy_parser import standard_transformations as _std_trf

    log.remark("Evaluating symbolic expression:  %s", expr)

    symbols = symbols if symbols else {}
    parse_kwargs = dict(
        evaluate=evaluate,
        transformations=(
            transformations if transformations is not None else _std_trf
        ),
    )

    # Now, parse the expression
    try:
        res = _parse_expr(expr, local_dict=symbols, **parse_kwargs)

    except Exception as exc:
        raise ValueError(
            f"Failed parsing expression '{expr}'! Got a "
            f"{exc.__class__.__name__}: {exc}. Check that the expression can "
            "be evaluated with the available symbols "
            f"({', '.join(symbols) if symbols else 'none specified'}) "
            "and inspect the chained exceptions for more information. Parse "
            f"arguments were: {parse_kwargs}"
        ) from exc

    # Finished here if no type cast is desired
    if astype is None:
        return res

    # If full evaluation is desired, do so via a numpy type cast. This works on
    # all sympy objects, but _importantly_ also works on numpy arrays
    # containing sympy objects.
    dtype = np.dtype(astype)
    log.debug("Applying %s ...", dtype)

    try:
        return dtype.type(res)

    except Exception as exc:
        raise TypeError(
            f"Failed casting the result of expression '{expr}' from "
            f"{type(res)} to {dtype}! This can also be due to free symbols "
            "remaining in the evaluated expression. Either specify the free "
            f"symbols (got: {', '.join(symbols) if symbols else 'none'}) or "
            "deactivate casting by specifying None as ``dtype`` argument. "
            f"The expression evaluated to:\n\n    {res}\n"
        ) from exc


def generate_lambda(expr: str) -> Callable:
    """Generates a lambda from a string. This is useful when working with
    callables in other operations.

    The ``expr`` argument needs to be a valid Python ``lambda`` expression, see
    `here <docs.python.org/3/tutorial/controlflow.html#lambda-expressions>`_.

    Inside the lambda body, the following names are available for use:

        * A large part of the ``builtins`` module
        * Every name from the Python ``math`` module, e.g. ``sin``, ``cos``, …
        * These modules (and their long form): ``np``, ``xr``, ``scipy``

    Internally, this uses ``eval`` but imposes the following restrictions:

        * The following strings may *not* appear in ``expr``: ``;``, ``__``.
        * There can be no nested ``lambda``, i.e. the only allowed lambda
          string is that in the beginning of ``expr``.
        * The dangerous parts from the ``builtins`` module are *not* available.

    Args:
        expr (str): The expression string to evaluate into a lambda.

    Returns:
        Callable: The generated Callable.

    Raises:
        SyntaxError: Upon failed evaluation of the given expression, invalid
            expression pattern, or disallowed strings in the lambda body.
    """
    ALLOWED_BUILTINS = (
        "abs",
        "all",
        "any",
        "callable",
        "chr",
        "divmod",
        "format",
        "hash",
        "hex",
        "id",
        "isinstance",
        "issubclass",
        "iter",
        "len",
        "max",
        "min",
        "next",
        "oct",
        "ord",
        "print",
        "repr",
        "round",
        "sorted",
        "sum",
        "None",
        "Ellipsis",
        "False",
        "True",
        "bool",
        "bytes",
        "bytearray",
        "complex",
        "dict",
        "enumerate",
        "filter",
        "float",
        "frozenset",
        "int",
        "list",
        "map",
        "range",
        "reversed",
        "set",
        "slice",
        "str",
        "tuple",
        "type",
        "zip",
        "open",
    )
    DISALLOWED_STRINGS = ("lambda", ";", "__")
    LAMBDA_PATTERN = r"^\s*lambda\s[\w\s\,\*]+\:(.+)$"
    #                   arguments definitions  : lambda body (capture group)
    # See also:  https://regex101.com/r/OmI8NY/2/

    # Check if the given expression matches this pattern
    pattern = re.compile(LAMBDA_PATTERN)
    match = pattern.match(expr)
    if match is None:
        raise SyntaxError(
            f"The given expression '{expr}' was not a valid lambda expression!"
        )

    # Check if the lambda body contains disallowed strings
    lambda_body = match[1]
    if any([bad_str in lambda_body for bad_str in DISALLOWED_STRINGS]):
        raise SyntaxError(
            "Encountered one or more disallowed strings in the "
            f"body ('{lambda_body}') of the given lambda "
            f"expression ('{expr}'). Make sure none of the "
            "following strings appears there: "
            f"{DISALLOWED_STRINGS}"
        )

    # Ok, sanitized enough now. Prepare the globals dict, restricting access to
    # a subset of builtins and only allowing commonly used math functionality
    _g = dict(
        __builtins__={
            n: f for n, f in __builtins__.items() if n in ALLOWED_BUILTINS
        },
        math=math,
        scipy=scipy,
        numpy=np,
        xarray=xr,
        np=np,
        xr=xr,
        **{n: f for n, f in math.__dict__.items() if not n.startswith("_")},
    )

    # Try evaluation now (with empty locals)
    try:
        f = eval(expr, _g, {})

    except Exception as exc:
        raise SyntaxError(
            "Failed generating lambda object from expression "
            f"'{expr}'! Got a {exc.__class__.__name__}: {exc}"
        ) from exc

    return f


# .............................................................................
# numpy and xarray operations


def create_mask(
    data: xr.DataArray, operator_name: str, rhs_value: float
) -> xr.DataArray:
    """Given the data, returns a binary mask by applying the following
    comparison: ``data <operator> rhs value``.

    Args:
        data (xr.DataArray): The data to apply the comparison to. This is the
            lhs of the comparison.
        operator_name (str): The name of the binary operator function as
            registered in the ``BOOLEAN_OPERATORS`` constant.
        rhs_value (float): The right-hand-side value

    Raises:
        KeyError: On invalid operator name

    Returns:
        xr.DataArray: Boolean mask
    """
    # Get the operator function
    try:
        comp_func = BOOLEAN_OPERATORS[operator_name]

    except KeyError as err:
        raise KeyError(
            f"No boolean operator '{operator_name}' available! Available "
            f"operators: {', '.join(BOOLEAN_OPERATORS.keys())}"
        ) from err

    # Apply the comparison
    data = comp_func(data, rhs_value)

    # If a name already exists, add some information about the masking to it
    if data.name:
        name = data.name + f" (masked by '{operator_name} {rhs_value}')"
    else:
        name = None

    # Build a new xr.DataArray from that data, retaining all information
    return xr.DataArray(
        data=data, name=name, dims=data.dims, coords=data.coords
    )


def where(
    data: xr.DataArray, operator_name: str, rhs_value: float
) -> xr.DataArray:
    """Filter elements from the given data according to a condition. Only
    those elemens where the condition is fulfilled are not masked.

    NOTE This leads to a dtype change to float.
    """
    # Get the mask and apply it
    return data.where(
        create_mask(data, operator_name=operator_name, rhs_value=rhs_value)
    )


def count_unique(data, dims: List[str] = None) -> xr.DataArray:
    """Applies np.unique to the given data and constructs a xr.DataArray for
    the results.

    NaN values are filtered out.

    Args:
        data: The data
        dims (List[str], optional): The dimensions along which to apply
            np.unique. The other dimensions will be available after the
            operation. If not provided it is applied along all dims.

    """

    def _count_unique(data) -> xr.DataArray:
        unique, counts = np.unique(data, return_counts=True)

        # remove np.nan values
        # NOTE np.nan != np.nan, hence np.nan will count 1 for every
        #      occurrence, but duplicate values are not allowed in coords...
        counts = counts[~np.isnan(unique)]
        unique = unique[~np.isnan(unique)]

        if isinstance(data, xr.DataArray) and data.name:
            name = data.name + " (unique counts)"
        else:
            name = "unique counts"

        # Construct a new data array and return
        return xr.DataArray(
            data=counts,
            name=name,
            dims=("unique",),
            coords=dict(unique=unique),
        )

    if not dims:
        return _count_unique(data)

    if not isinstance(data, xr.DataArray):
        raise TypeError(
            f"Data needs to be of type xr.DataArray, but was {type(data)}!"
        )

    # use split-apply-combine along those dimensions not in dims
    split_dims = [dim for dim in data.dims if dim not in dims]

    if len(split_dims) == 0:
        return _count_unique(data)

    data = data.stack(_stack_cu=split_dims).groupby("_stack_cu")
    return data.map(_count_unique).unstack("_stack_cu")


# .............................................................................
# Working with multidimensional data, mostly xarray-based


def populate_ndarray(
    objs: Iterable,
    shape: Tuple[int] = None,
    dtype: Union[str, type, np.dtype] = float,
    order: str = "C",
    out: np.ndarray = None,
    ufunc: Callable = None,
) -> np.ndarray:
    """Populates an empty np.ndarray of the given dtype with the given objects
    by zipping over a new array of the given ``shape`` and the sequence of
    objects.

    Args:
        objs (Iterable): The objects to add to the np.ndarray. These objects
            are added in the order they are given here. Note that their final
            position inside the resulting array is furthermore determined by
            the ``order`` argument.
        shape (Tuple[int], optional): The shape of the new array. **Required**
            if no ``out`` array is given.
        dtype (Union[str, type, np.dtype], optional): dtype of the new array.
            Ignored if ``out`` is given.
        order (str, optional): Order of the new array, determines iteration
            order. Ignored if ``out`` is given.
        out (np.ndarray, optional): If given, populates this array rather than
            an empty array.
        ufunc (Callable, optional): If given, applies this unary function to
            each element before storing it in the to-be-returned ndarray.

    Returns:
        np.ndarray: The populated ``out`` array or the newly created one (if
            ``out`` was not given)

    Raises:
        TypeError: On missing
        ValueError: If the number of given objects did not match the array size
    """
    if out is None and shape is None:
        raise TypeError(
            "Without an output array given, the `shape` argument "
            "needs to be specified!"
        )

    ufunc = ufunc if ufunc is not None else lambda e: e
    out = out if out is not None else np.empty(shape, dtype=dtype, order=order)

    if len(objs) != out.size:
        raise ValueError(
            f"Mismatch between array size ({out.size}, shape: {out.shape}) "
            f"and number of given objects ({len(objs)})!"
        )

    it = np.nditer(out, flags=("multi_index", "refs_ok"))
    for obj, _ in zip(objs, it):
        out[it.multi_index] = ufunc(obj)

    return out


def multi_concat(arrs: np.ndarray, *, dims: Sequence[str]) -> xr.DataArray:
    """Concatenates ``xr.Dataset`` or ``xr.DataArray`` objects using
    ``xr.concat``. This function expects the xarray objects to be pre-aligned
    inside the numpy *object* array ``arrs``, with the number of dimensions
    matching the number of concatenation operations desired.
    The position inside the array carries information on where the objects that
    are to be concatenated are placed inside the higher dimensional coordinate
    system.

    Through multiple concatenation, the dimensionality of the contained objects
    is increased by ``dims``, while their dtype can be maintained.

    For the sequential application of ``xr.concat`` along the outer dimensions,
    the custom :py:func:`dantro.tools.apply_along_axis` is used.

    Args:
        arrs (np.ndarray): The array containing xarray objects which are to be
            concatenated. Each array dimension should correspond to one of the
            given ``dims``. For each of the dimensions, the ``xr.concat``
            operation is applied along the axis, effectively reducing the
            dimensionality of ``arrs`` to a scalar and increasing the
            dimensionality of the contained xarray objects until they
            additionally contain the dimensions specified in ``dims``.
        dims (Sequence[str]): A sequence of dimension names that is assumed to
            match the dimension names of the array. During each concatenation
            operation, the name is passed along to ``xr.concat`` where it is
            used to select the dimension of the *content* of ``arrs`` along
            which concatenation should occur.

    Raises:
        ValueError: If number of dimension names does not match the number of
            data dimensions.
    """
    # Check dimensionality
    if len(dims) != arrs.ndim:
        raise ValueError(
            f"The given sequence of dimension names, {dims}, did not match "
            f"the number of dimensions of data of shape {arrs.shape}!"
        )

    # Reverse-iterate over dimensions and concatenate them
    for dim_idx, dim_name in reversed(list(enumerate(dims))):
        log.debug(
            "Concatenating along axis '%s' (idx: %d) ...", dim_name, dim_idx
        )

        arrs = apply_along_axis(
            xr.concat, axis=dim_idx, arr=arrs, dim=dim_name
        )
        # NOTE ``np.apply_along_axis`` would be what is desired here, but that
        #      function unfortunately tries to cast objects to np.arrays which
        #      is not what we want here at all! Thus, this function uses the
        #      custom dantro function of the same name instead.

    # Should be scalar now, get the element.
    return arrs.item()


def merge(
    arrs: Union[Sequence[Union[xr.DataArray, xr.Dataset]], np.ndarray],
    *,
    reduce_to_array: bool = False,
    **merge_kwargs,
) -> Union[xr.Dataset, xr.DataArray]:
    """Merges the given sequence of xarray objects into an xr.Dataset.

    As a convenience, this also allows passing a numpy object array containing
    the xarray objects. Furthermore, if the resulting Dataset contains only a
    single data variable, that variable can be extracted as a DataArray which
    is then the return value of this operation.
    """
    if isinstance(arrs, np.ndarray):
        arrs = arrs.flat

    dset = xr.merge(arrs, **merge_kwargs)

    if not reduce_to_array:
        return dset

    elif len(dset.data_vars) != 1:
        raise ValueError(
            "Can only reduce the Dataset resulting from the xr.merge "
            "operation to a DataArray if one and only one data variable is "
            "present in the Dataset! "
            f"Got: {', '.join(dset.data_vars)}. Full data:\n{dset}"
        )

    # Get the name of the single data variable and then get the DataArray
    darr = dset[list(dset.data_vars.keys())[0]]
    # NOTE This is something else than the Dataset.to_array() method, which
    #      includes the name of the data variable as another coordinate. This
    #      is not desired, because it is not relevant.
    return darr


def expand_dims(
    d: Union[np.ndarray, xr.DataArray], *, dim: dict = None, **kwargs
) -> xr.DataArray:
    """Expands the dimensions of the given object.

    If the object does not support the ``expand_dims`` method, it will be
    attempted to convert it to an xr.DataArray.

    Args:
        d (Union[np.ndarray, xr.DataArray]): The object to expand the
            dimensions of
        dim (dict, optional): Keys specify the dimensions to expand, values can
            either be an integer specifying the length of the dimension, or a
            sequence of coordinates.
        **kwargs: Passed on to ``expand_dims`` method

    Returns:
        xr.DataArray: The input data with expanded dimensions.
    """
    if not hasattr(d, "expand_dims"):
        d = xr.DataArray(d)
    return d.expand_dims(dim, **kwargs)


def expand_object_array(
    d: xr.DataArray,
    *,
    shape: Sequence[int] = None,
    astype: Union[str, type, np.dtype] = None,
    dims: Sequence[str] = None,
    coords: Union[dict, str] = "trivial",
    combination_method: str = "concat",
    allow_reshaping_failure: bool = False,
    **combination_kwargs,
) -> xr.DataArray:
    """Expands a labelled object-array that contains array-like objects into a
    higher-dimensional labelled array.

    ``d`` is expected to be an array *of arrays*, i.e. each element of the
    outer array is an object that itself is an ``np.ndarray``-like object.
    The ``shape`` is the expected shape of each of these *inner* arrays.
    Importantly, all these arrays need to have the exact same shape.

    Typically, e.g. when loading data from HDF5 files, the inner array will
    not be labelled but will consist of simple np.ndarrays.
    The arguments ``dims`` and ``coords`` are used to label the *inner* arrays.

    This uses :py:func:`~dantro.utils.data_ops.multi_concat` for concatenating
    or :py:func:`~dantro.utils.data_ops.merge` for merging the object arrays
    into a higher-dimensional array, where the latter option allows for missing
    values.

    .. TODO::

        Make reshaping and labelling optional if the inner array already is a
        labelled array. In such cases, the coordinate assignment is already
        done and all information for combination is already available.

    Args:
        d (xr.DataArray): The labelled object-array containing further arrays
            as elements (which are assumed to be unlabelled).
        shape (Sequence[int], optional): Shape of the inner arrays. If not
            given, the first element is used to determine the shape.
        astype (Union[str, type, np.dtype], optional): All inner arrays need to
            have the same dtype. If this argument is given, the arrays will be
            coerced to this dtype. For numeric data, ``float`` is typically a
            good fallback. Note that with ``combination_method == "merge"``,
            the choice here might not be respected.
        dims (Sequence[str], optional): Dimension names for labelling the
            inner arrays. This is necessary for proper alignment. The number of
            dimensions need to match the ``shape``. If not given, will use
            ``inner_dim_0`` and so on.
        coords (Union[dict, str], optional): Coordinates of the inner arrays.
            These are necessary to align the inner arrays with each other. With
            ``coords = "trivial"``, trivial coordinates will be assigned to all
            dimensions. If specifying a dict and giving ``"trivial"`` as value,
            that dimension will be assigned trivial coordinates.
        combination_method (str, optional): The combination method to use to
            combine the object array. For ``concat``, will use dantro's
            :py:func:`~dantro.utils.data_ops.multi_concat`, which preserves
            dtype but does not allow missing values. For ``merge``, will use
            :py:func:`~dantro.utils.data_ops.merge`, which allows missing
            values (masked using ``np.nan``) but leads to the dtype decaying
            to float.
        allow_reshaping_failure (bool, optional): If true, the expansion is not
            stopped if reshaping to ``shape`` fails for an element. This will
            lead to missing values at the respective coordinates and the
            ``combination_method`` will automatically be changed to ``merge``.
        **combination_kwargs: Passed on to the selected combination function,
            :py:func:`~dantro.utils.data_ops.multi_concat` or
            :py:func:`~dantro.utils.data_ops.merge`.

    Returns:
        xr.DataArray: A new, higher-dimensional labelled array.

    Raises:
        TypeError: If no ``shape`` can be extracted from the first element in
            the input data ``d``
        ValueError: On bad argument values for ``dims``, ``shape``, ``coords``
            or ``combination_method``.
    """

    def prepare_item(
        d: xr.DataArray,
        *,
        midx: Sequence[int],
        shape: Sequence[int],
        astype: Union[str, type, np.dtype, None],
        name: str,
        dims: Sequence[str],
        generate_coords: Callable,
    ) -> Union[xr.DataArray, None]:
        """Extracts the desired element and reshapes and labels it accordingly.
        If any of this fails, returns ``None``.
        """
        elem = d[midx]

        try:
            item = elem.item().reshape(shape)

        except Exception as exc:
            if allow_reshaping_failure:
                return None
            raise ValueError(
                f"Failed reshaping item at {midx} to {shape}! Make sure the "
                f"element\n\n{elem}\n\nallows reshaping. To discard values "
                "where reshaping fails, enable `allow_reshaping_failure`."
            ) from exc

        if astype is not None:
            item = item.astype(astype)

        return xr.DataArray(
            item, name=name, dims=dims, coords=generate_coords(elem)
        )

    # Make sure we are operating on labelled data
    d = xr.DataArray(d)

    # Try to deduce missing arguments and make sure arguments are ok
    if shape is None:
        try:
            shape = d.data.flat[0].shape
        except Exception as exc:
            raise TypeError(
                "Failed extracting a shape from the first element of the "
                f"given array:\n{d}\nCheck that the given array contains "
                "further np.ndarray-like objects. Alternatively, explicitly "
                "provide the `shape` argument."
            ) from exc

    if dims is None:
        dims = tuple([f"inner_dim_{n:d}" for n, _ in enumerate(shape)])

    if len(dims) != len(shape):
        raise ValueError(
            "Number of dimension names and number of dimensions of the inner "
            f"arrays needs to match! Got dimension names {dims} for array "
            f"elements of expected shape {shape}."
        )

    if coords == "trivial":
        coords = {n: "trivial" for n in dims}

    elif not isinstance(coords, dict):
        raise TypeError(
            f"Argument `coords` needs to be a dict or str, but was {coords}!"
        )

    if set(coords.keys()) != set(dims):
        raise ValueError(
            "Mismatch between dimension names and coordinate keys! Make sure "
            "there are coordinates specified for each dimension of the inner "
            f"arrays, {dims}! Got:\n{coords}"
        )

    # Handle trivial coordinates for each coordinate dimension separately
    coords = {
        n: (range(l) if isinstance(c, str) and c == "trivial" else c)
        for (n, c), l in zip(coords.items(), shape)
    }

    # Assemble info needed to bring individual array items into proper form
    item_name = d.name if d.name else "data"
    item_shape = tuple([1 for _ in d.shape]) + tuple(shape)
    item_dims = d.dims + tuple(dims)
    item_coords = lambda e: dict(
        **{n: [c.item()] for n, c in e.coords.items()}, **coords
    )

    # The array that gathers all to-be-combined object arrays
    arrs = np.empty_like(d, dtype=object)
    arrs.fill(dict())  # are ignored in xr.merge

    # Transform each element to a labelled xr.DataArray that includes the outer
    # dimensions and coordinates; the latter is crucial for alignment.
    # Alongside, type coercion can be performed. For failed reshaping, the
    # element may be skipped.
    it = np.nditer(arrs.data, flags=("multi_index", "refs_ok"))
    for _ in it:
        item = prepare_item(
            d,
            midx=it.multi_index,
            shape=item_shape,
            astype=astype,
            name=item_name,
            dims=item_dims,
            generate_coords=item_coords,
        )

        if item is None:
            # Missing value; need to fall back to combination via merge
            combination_method = "merge"
            continue
        arrs[it.multi_index] = item

    # Now, combine
    if combination_method == "concat":
        return multi_concat(arrs, dims=d.dims, **combination_kwargs)

    elif combination_method == "merge":
        return merge(arrs, reduce_to_array=True, **combination_kwargs)

    raise ValueError(
        f"Invalid combination method '{combination_method}'! "
        "Choose from: 'concat', 'merge'."
    )


# fmt: off
# -----------------------------------------------------------------------------
# The Operations Database -----------------------------------------------------
# NOTE If a single "object to act upon" can be reasonably defined for an
#      operation, it should be accepted as the first positional argument.
_OPERATIONS = KeyOrderedDict({
    # General operations - - - - - - - - - - - - - - - - - - - - - - - - - - -
    'define':       lambda d: d,
    'pass':         lambda d: d,
    'print':        print_data,

    # Working on imported modules (useful if other operations don't match)
    'from_module':  get_from_module,
    'import':       import_module_or_object,
    'call':         lambda c, *a, **k: c(*a, **k),
    'import_and_call':
        lambda m, n, *a, **k: import_module_or_object(m, n)(*a, **k),

    # Import from packages and call directly
    'np.':
        lambda ms, *a, **k: get_from_module(np, name=ms)(*a, **k),
    'xr.':
        lambda ms, *a, **k: get_from_module(xr, name=ms)(*a, **k),
    'scipy.':
        lambda ms, *a, **k: get_from_module(scipy, name=ms)(*a, **k),
    'nx.':
        lambda ms, *a, **k: get_from_module(nx, name=ms)(*a, **k),

    # Defining lambdas
    'lambda':       generate_lambda,
    'call_lambda':  lambda e, *a, **k: generate_lambda(e)(*a, **k),

    # Some commonly used types
    'list':         list,
    'dict':         dict,
    'tuple':        tuple,
    'set':          set,

    'int':          int,
    'float':        float,
    'complex':      complex,
    'bool':         bool,
    'str':          str,

    # Item access and manipulation
    'getitem':              lambda d, k:    d[k],
    'setitem':              lambda d, k, v: d.__setitem__(k, v),
    'recursive_getitem':    recursive_getitem,

    # Attribute-related
    'getattr':      getattr,
    'setattr':      setattr,
    'callattr':     lambda d, attr, *a, **k: getattr(d, attr)(*a, **k),

    # Other common Python builtins
    'all':          all,
    'any':          any,
    'len':          len,
    'min':          min,
    'max':          max,
    'sum':          sum,
    'map':          map,
    'repr':         repr,
    'sorted':       sorted,

    # Common operations on strings
    '.format':      lambda s, *a, **k: s.format(*a, **k),
    '.join':        lambda s, *a, **k: s.join(*a, **k),
    '.split':       lambda s, *a, **k: s.split(*a, **k),


    # Numerical operations - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Unary ...................................................................
    'neg':          lambda d: operator.neg(d),
    'pos':          lambda d: operator.pos(d),
    'truth':        lambda d: operator.truth(d),

    'increment':    lambda d: d + 1,
    'decrement':    lambda d: d - 1,
    'count_unique': count_unique,

    # numpy
    '.T':           lambda d: d.T,
    '.any':         lambda d: d.any(),
    '.all':         lambda d: d.all(),
    '.dtype':       lambda d: d.dtype,
    '.shape':       lambda d: d.shape,
    '.ndim':        lambda d: d.ndim,
    '.size':        lambda d: d.size,
    '.itemsize':    lambda d: d.itemsize,
    '.nbytes':      lambda d: d.nbytes,
    '.base':        lambda d: d.base,
    '.imag':        lambda d: d.imag,
    '.real':        lambda d: d.real,
    '.nonzero':     lambda d: d.nonzero,
    '.flat':        lambda d: d.flat,
    '.item':        lambda d: d.item(),

    # xarray
    '.head':        lambda d: d.head(),
    '.tail':        lambda d: d.tail(),

    # logarithms and powers
    'log':          lambda d: np.log(d),
    'log10':        lambda d: np.log10(d),
    'log2':         lambda d: np.log2(d),
    'log1p':        lambda d: np.log1p(d),
    'squared':      lambda d: np.square(d),
    'sqrt':         lambda d: np.sqrt(d),
    'cubed':        lambda d: np.power(d, 3),
    'sqrt3':        lambda d: np.power(d, 1./.3),

    # normalization and cumulation
    'normalize_to_sum':         lambda d: d / np.sum(d),
    'normalize_to_max':         lambda d: d / np.max(d),
    'cumulate':                 lambda d: np.cumsum(d),
    'cumulate_complementary':   lambda d: np.cumsum(d[::-1])[::-1],


    # Binary ..................................................................
    # Elementwise operations
    'add':          lambda d, v: operator.add(d, v),
    'concat':       lambda d, v: operator.concat(d, v),
    'div':          lambda d, v: operator.truediv(d, v),
    'truediv':      lambda d, v: operator.truediv(d, v),
    'floordiv':     lambda d, v: operator.floordiv(d, v),
    'lshift':       lambda d, v: operator.lshift(d, v),
    'mod':          lambda d, v: operator.mod(d, v),
    'mul':          lambda d, v: operator.mul(d, v),
    'matmul':       lambda d, v: operator.matmul(d, v),
    'pow':          lambda d, v: operator.pow(d, v),
    'rshift':       lambda d, v: operator.rshift(d, v),
    'sub':          lambda d, v: operator.sub(d, v),

    # numpy
    'power':        lambda d, e: np.power(d, e),
    'np.dot':       np.dot,

    # xarray
    '.coords':      lambda d, key: d.coords[key],


    # N-ary ...................................................................
    # Masking, array generation
    'create_mask':                  create_mask,
    'where':                        where,
    'populate_ndarray':             populate_ndarray,
    'expand_object_array':          expand_object_array,

    # extract labelling info, e.g. for creating higher-dimensional arrays
    'extract_dim_names':            extract_dim_names,
    'extract_coords_from_attrs':    extract_coords_from_attrs,

    # dantro-specific wrappers around other library's functionality
    'dantro.apply_along_axis':      apply_along_axis,
    'dantro.multi_concat':          multi_concat,
    'dantro.merge':                 merge,
    'dantro.expand_dims':           expand_dims,

    # evaluating symbolic expressions using sympy
    'expression':                   expression,
    # NOTE: The `^` operator acts as XOR; use `**` for exponentiation!

    # numpy
    '.sum':             lambda d, *a, **k: d.sum(*a, **k),
    '.prod':            lambda d, *a, **k: d.prod(*a, **k),
    '.cumsum':          lambda d, *a, **k: d.cumsum(*a, **k),
    '.cumprod':         lambda d, *a, **k: d.cumprod(*a, **k),

    '.mean':            lambda d, *a, **k: d.mean(*a, **k),
    '.std':             lambda d, *a, **k: d.std(*a, **k),
    '.min':             lambda d, *a, **k: d.min(*a, **k),
    '.max':             lambda d, *a, **k: d.max(*a, **k),
    '.var':             lambda d, *a, **k: d.var(*a, **k),
    '.argmin':          lambda d, *a, **k: d.argmin(*a, **k),
    '.argmax':          lambda d, *a, **k: d.argmax(*a, **k),
    '.argsort':         lambda d, *a, **k: d.argsort(*a, **k),
    '.argpartition':    lambda d, *a, **k: d.argpartition(*a, **k),

    '.transpose':       lambda d, *ax: d.transpose(*ax),
    '.squeeze':         lambda d, **k: d.squeeze(**k),
    '.flatten':         lambda d, **k: d.flatten(**k),
    '.diagonal':        lambda d, **k: d.diagonal(**k),
    '.trace':           lambda d, **k: d.trace(**k),
    '.sort':            lambda d, **k: d.sort(**k),
    '.fill':            lambda d, val: d.fill(val),
    '.round':           lambda d, **k: d.round(**k),
    '.take':            lambda d, i, **k: d.take(i, **k),
    '.swapaxes':        lambda d, a1, a2: d.swapaxes(a1, a2),
    '.reshape':         lambda d, s, **k: d.reshape(s, **k),
    '.astype':          lambda d, t, **k: d.astype(t, **k),

    'np.array':         np.array,
    'np.empty':         np.empty,
    'np.zeros':         np.zeros,
    'np.ones':          np.ones,
    'np.empty_like':    np.empty_like,
    'np.zeros_like':    np.zeros_like,
    'np.ones_like':     np.ones_like,

    'np.eye':           np.eye,
    'np.arange':        np.arange,
    'np.linspace':      np.linspace,
    'np.logspace':      np.logspace,

    'np.invert':        np.invert,
    'np.transpose':     np.transpose,
    'np.flip':          np.flip,
    'np.diff':          np.diff,
    'np.reshape':       np.reshape,
    'np.take':          np.take,
    'np.repeat':        np.repeat,
    'np.stack':         np.stack,
    'np.hstack':        np.hstack,
    'np.vstack':        np.vstack,
    'np.concatenate':   np.concatenate,

    'np.abs':           np.abs,
    'np.ceil':          np.ceil,
    'np.floor':         np.floor,
    'np.round':         np.round,

    'np.where':         np.where,
    'np.digitize':      np.digitize,
    'np.histogram':     np.histogram,
    'np.count_nonzero': np.count_nonzero,

    # xarray
    '.sel':             lambda d, *a, **k: d.sel(*a, **k),
    '.isel':            lambda d, *a, **k: d.isel(*a, **k),
    '.drop_sel':        lambda d, *a, **k: d.drop_sel(*a, **k),
    '.median':          lambda d, *a, **k: d.median(*a, **k),
    '.quantile':        lambda d, *a, **k: d.quantile(*a, **k),
    '.argmin':          lambda d, *a, **k: d.argmin(*a, **k),
    '.argmax':          lambda d, *a, **k: d.argmax(*a, **k),
    '.count':           lambda d, *a, **k: d.count(*a, **k),
    '.diff':            lambda d, *a, **k: d.diff(*a, **k),
    '.where':           lambda d, c, *a, **k: d.where(c, *a, **k),
    '.ffill':           lambda d, *a, **k: d.ffill(*a, **k),
    '.bfill':           lambda d, *a, **k: d.bfill(*a, **k),
    '.fillna':          lambda d, *a, **k: d.fillna(*a, **k),
    '.interpolate_na':  lambda d, *a, **k: d.interpolate_na(*a, **k),
    '.dropna':          lambda d, *a, **k: d.dropna(*a, **k),
    '.isin':            lambda d, *a, **k: d.isin(*a, **k),

    '.groupby':         lambda d, g, **k: d.groupby(g, **k),
    '.groupby_bins':    lambda d, g, **k: d.groupby_bins(g, **k),
    '.map':             lambda ds, func, **k: ds.map(func, **k),
    '.reduce':          lambda d, func, **k: d.reduce(func, **k),

    '.rename':          lambda d, *a, **k: d.rename(*a, **k),
    '.expand_dims':     lambda d, *a, **k: d.expand_dims(*a, **k),
    '.assign_coords':   lambda d, *a, **k: d.assign_coords(*a, **k),
    '.assign_attrs':    lambda d, *a, **k: d.assign_attrs(*a, **k),
    '.assign':          lambda d, *a, **k: d.assign(*a, **k),

    '.to_array':        lambda ds, *a, **k: ds.to_array(*a, **k),

    '.to_dataframe':    lambda d, *a, **k: d.to_dataframe(*a, **k),

    'xr.Dataset':       xr.Dataset,
    'xr.DataArray':     xr.DataArray,
    'xr.zeros_like':    xr.zeros_like,
    'xr.ones_like':     xr.ones_like,

    'xr.merge':             xr.merge,
    'xr.concat':            xr.concat,
    'xr.align':             xr.align,
    'xr.combine_nested':    xr.combine_nested,
    'xr.combine_by_coords': xr.combine_by_coords,

    # scipy
    'curve_fit':
        lambda *a, **k: import_module_or_object("scipy.optimize",
                                                name="curve_fit")(*a, **k),
    # NOTE: Use the 'lambda' operation to generate the callable
}) # End of default operation definitions
# fmt: on

# Add the boolean operators
_OPERATIONS.update(BOOLEAN_OPERATORS)


# -----------------------------------------------------------------------------
# Registering and applying operations


def register_operation(
    *,
    name: str,
    func: Callable,
    skip_existing: bool = False,
    overwrite_existing: bool = False,
) -> None:
    """Adds an entry to the shared operations registry.

    Args:
        name (str): The name of the operation
        func (Callable): The callable
        skip_existing (bool, optional): Whether to skip registration if the
            operation name is already registered. This suppresses the
            ValueError raised on existing operation name.
        overwrite_existing (bool, optional): Whether to overwrite a potentially
            already existing operation of the same name. If given, this takes
            precedence over ``skip_existing``.

    Raises:
        TypeError: On invalid name or non-callable for the func argument
        ValueError: On already existing operation name and no skipping or
            overwriting enabled.
    """
    if name in _OPERATIONS and not overwrite_existing:
        if skip_existing:
            log.debug(
                "Operation '%s' is already registered and will not be "
                "registered again.",
                name,
            )
            return
        raise ValueError(
            f"Operation name '{name}' already exists! Refusing to register a "
            "new one. Set the overwrite_existing flag to force overwriting."
        )

    elif not callable(func):
        raise TypeError(
            f"The given {func} for operation '{name}' is not callable! "
        )

    elif not isinstance(name, str):
        raise TypeError(
            f"Operation name need be a string, was {type(name)} with "
            f"value {name}!"
        )

    _OPERATIONS[name] = func
    log.debug("Registered operation '%s'.", name)


def apply_operation(
    op_name: str, *op_args, _log_level: int = 5, **op_kwargs
) -> Any:
    """Apply an operation with the given arguments and then return it.

    Args:
        op_name (str): The name of the operation to carry out; need to be part
            of the OPERATIONS database.
        *op_args: The positional arguments to the operation
        _log_level (int, optional): Log level of the log messages created by
            this function.
        **op_kwargs: The keyword arguments to the operation

    Returns:
        Any: The result of the operation

    Raises:
        ValueError: On invalid operation name. This also suggests possible
            other names that might match.
        RuntimeError: On failure to *apply* the operation.
    """
    try:
        op = _OPERATIONS[op_name]

    except KeyError as err:
        # Find some close matches to make operation discovery easier
        _close_matches = available_operations(match=op_name, n=8)
        _did_you_mean = (
            f" Did you mean: {', '.join(_close_matches)} ?"
            if _close_matches
            else ""
        )
        _available = make_columns(available_operations())

        raise ValueError(
            f"No operation '{op_name}' registered!{_did_you_mean} "
            f"\nAvailable operations:\n{_available}If you need to register "
            "a new operation, use dantro.utils.register_operation."
        ) from err

    # Compute and return the results
    log.log(_log_level, "Performing operation '%s' ...", op_name)
    try:
        return op(*op_args, **op_kwargs)

    except Exception as exc:
        raise RuntimeError(
            f"Failed applying operation '{op_name}'! "
            f"Got a {exc.__class__.__name__}: {exc}\n"
            f"  args:   {op_args}\n"
            f"  kwargs: {op_kwargs}\n"
        ) from exc


def available_operations(*, match: str = None, n: int = 5) -> Sequence[str]:
    """Returns all available operation names or a fuzzy-matched subset of them.

    Args:
        match (str, optional): If given, fuzzy-matches the names and only
            returns close matches to this name.
        n (int, optional): Number of close matches to return. Passed on to
            difflib.get_close_matches

    Returns:
        Sequence[str]: All available operation names or the matched subset.
            The sequence is sorted alphabetically.
    """
    if match is None:
        return _OPERATIONS.keys()

    # Use fuzzy matching to return close matches
    return _get_close_matches(match, _OPERATIONS.keys(), n=n)



"""This module implements the DeclarativePlotCreator class"""

import logging

from .pcr_base import BasePlotCreator

# Local constants
log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


class DeclarativePlotCreator(BasePlotCreator):
    """This PlotCreator can create plots from a dantro-specific declarative
    plot configuration. The language is inspired by Vega-Lite but is adapted to
    working with the different data structures stored in a DataManager.
    """

    pass



"""Generic, DAG-based multiplot function for the
:py:class:`~dantro.plot_creators.pcr_ext.ExternalPlotCreator` and derived plot
creators.
"""

import copy
import logging
from typing import Callable, Union

import matplotlib.pyplot as plt

from ..._import_tools import LazyLoader
from ...exceptions import PlottingError
from ...tools import make_columns, recursive_update
from ..pcr_ext import PlotHelper, is_plot_func

# Local constants
log = logging.getLogger(__name__)

# Lazy module loading for packages that take a long time to import
_sns = LazyLoader("seaborn", _depth=1)
# NOTE This import is specifically for the definition of the seaborn-based
#      multiplot functions and should not be used for anything else!
#      Depth 1 means: `_sns.foobar` is still a LazyLoader object and will only
#      be resolved upon an *attribute* call! Subsequently, these need to be
#      resolved manually, as done via `_resolve_lazy_imports`.

# fmt: off

# The available plot kinds for the multiplot interface.
# Details of the seaborn-related plots can be found here in the seaborn API:
# https://seaborn.pydata.org/api.html
#
# NOTE Seaborn plot functions are defined here in a lazy fashion, thus being
#      actually LazyLoader instances. They are resolved upon a first call to
#      the multiplot function
_MULTIPLOT_FUNC_KINDS = { # --- start literalinclude
    # Seaborn - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # https://seaborn.pydata.org/api.html

    # Relational plots
    "sns.scatterplot":      _sns.scatterplot,
    "sns.lineplot":         _sns.lineplot,

    # Distribution plots
    "sns.histplot":         _sns.histplot,
    "sns.kdeplot":          _sns.kdeplot,
    "sns.ecdfplot":         _sns.ecdfplot,
    "sns.rugplot":          _sns.rugplot,

    # Categorical plots
    "sns.stripplot":        _sns.stripplot,
    "sns.swarmplot":        _sns.swarmplot,
    "sns.boxplot":          _sns.boxplot,
    "sns.violinplot":       _sns.violinplot,
    "sns.boxenplot":        _sns.boxenplot,
    "sns.pointplot":        _sns.pointplot,
    "sns.barplot":          _sns.barplot,
    "sns.countplot":        _sns.countplot,

    # Regression plots
    "sns.regplot":          _sns.regplot,
    "sns.residplot":        _sns.residplot,

    # Matrix plots
    "sns.heatmap":          _sns.heatmap,

    # Utility functions
    "sns.despine":          _sns.despine,

    # Matplotlib - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # https://matplotlib.org/tutorials/introductory/sample_plots.html

    # Relational plots
    "plt.fill":             plt.fill,
    "plt.scatter":          plt.scatter,
    "plt.plot":             plt.plot,
    "plt.polar":            plt.polar,
    "plt.loglog":           plt.loglog,
    "plt.semilogx":         plt.fill,
    "plt.semilogy":         plt.semilogy,

    # Distribution plots
    "plt.hist":             plt.hist,
    "plt.hist2d":           plt.hist2d,

    # Categorical plots
    "plt.bar":              plt.bar,
    "plt.barh":             plt.barh,
    "plt.pie":              plt.pie,
    "plt.table":            plt.table,

    # Matrix plots
    "plt.imshow":           plt.imshow,
    "plt.pcolormesh":       plt.pcolormesh,

    # Vector plots
    "plt.contour":          plt.contour,
    "plt.quiver":           plt.quiver,
    "plt.streamplot":       plt.streamplot,
}   # --- end literalinclude

# The multiplot functions that emit a warning if they do not get any arguments
# when called.
# This is helpful for functions that e.g. require a 'data' argument but do
# not fail or warn if no 'data' is passed on to them.
_MULTIPLOT_CAUTION_FUNC_NAMES = tuple([
    func_name for func_name in _MULTIPLOT_FUNC_KINDS
    if func_name not in ("sns.despine",)
])


# fmt: on


# -- Helper functions ---------------------------------------------------------


def _resolve_lazy_imports(d: dict):
    """In-place resolves lazy imports in the given dict"""
    for k, v in d.items():
        if isinstance(v, LazyLoader):
            d[k] = v.resolve()


def _parse_func_kwargs(
    function: Union[str, Callable],
    args: list = None,
    shared_kwargs: dict = None,
    **func_kwargs,
):
    """Parse a multiplot callable and its positional and keyword arguments.
    If ``function`` is a string it is looked up and mapped from the following
    dictionary:

    .. literalinclude:: ../../dantro/plot_creators/ext_funcs/multiplot.py
        :language: python
        :start-after: _MULTIPLOT_FUNC_KINDS = { # --- start literalinclude
        :end-before:  }   # --- end literalinclude
        :dedent: 4

    Args:
        function (Union[str, Callable]):  The callable function object or the
            name of the plot function to look up.

        args (list, optional): The positional arguments for the plot function

        shared_kwargs (dict, optional): Shared kwargs that passed on to
            all multiplot functions. They are recursively updated with
            the individual plot functions' func_kwargs.

        **func_kwargs (dict): The function kwargs to be passed on to the
            function object.

    .. note::

        The function kwargs cannot pass on a ``function`` or ``args`` key
        because both are parsed and translated into the plot function to use
        and the optional positional function arguments, respectively.

    Returns:
        (str, Callable, list, dict): (function name, function object, function
            arguments, function kwargs)
    """
    # First need to resolve all lazy imports in _MULTIPLOT_FUNC_KINDS
    _resolve_lazy_imports(_MULTIPLOT_FUNC_KINDS)

    if shared_kwargs is None:
        shared_kwargs = {}

    func_kwargs = recursive_update(copy.deepcopy(shared_kwargs), func_kwargs)

    if callable(function):
        func_name = function.__name__
        func = function
    else:
        func_name = function

        # Look up the function in the _MULTIPLOT_FUNC_KINDS dict
        try:
            func = _MULTIPLOT_FUNC_KINDS[func_name]
        except KeyError as err:
            _mp_funcs = _MULTIPLOT_FUNC_KINDS
            if _mp_funcs:
                _mp_funcs = "\n" + make_columns(_MULTIPLOT_FUNC_KINDS)
            else:
                _mp_funcs = " (none)\n"

            raise ValueError(
                f"The function `{func_name}` is not a valid multiplot "
                f"function. Available functions: \n {_mp_funcs} \n"
                "Alternatively, pass a callable instead of the name of a plot "
                "function."
            ) from err

    if args is None:
        args = []

    return func_name, func, args, func_kwargs


# -----------------------------------------------------------------------------
# -- The actual plotting functions --------------------------------------------
# -----------------------------------------------------------------------------


@is_plot_func(use_dag=True)
def multiplot(
    *,
    hlpr: PlotHelper,
    to_plot: Union[list, dict],
    data: dict,
    show_hints: bool = True,
    **shared_kwargs,
) -> None:
    """Consecutively plot multiple functions on one or multiple axes.

    ``to_plot`` contains all relevant information for the functions to plot.
    If ``to_plot`` is list-like the plot functions are plotted on the current
    axes created through the hlpr.
    If ``to_plot`` is dict-like, the keys specify the coordinate pair selecting
    an ax to plot on, e.g. (0,0), while the values specify a list of plot
    function configurations to apply consecutively.
    Each list entry specifies one function plot and is parsed via the
    :py:func:`~dantro.plot_creators.ext_funcs.multiplot._parse_func_kwargs`
    function.
    The multiplot works with any plot function that either operates on the
    current axis and does _not_ create a new figure or does not require an
    axis at all.

    Look at the :ref:`multiplot documentation <dag_multiplot>` for further
    information.

    Examples:
        A simple ``to_plot`` configuration looks like this:

        .. code-block:: yaml

            to_plot:
              - function: sns.lineplot
                data: !dag_result data
                # Note that especially seaborn plot functions require a
                # `data` input argument that can conveniently be
                # provided via the `!dag_result` YAML-tag.
                # If not provided, nothing is plotted without emitting
                # a warning.
              - function: sns.despine

        A simple ``to_plot`` configuration specifying two axis is:

        .. code-block:: yaml

            to_plot:
              [0,0]:
                - function: sns.lineplot
                  data: !dag_result data
              [1,0]:
                - function: sns.scatterplot
                  data: !dag_result data

    Args:
        hlpr (PlotHelper): The PlotHelper instance for this plot
        to_plot (Union[list, dict]): The data to plot.
        data (dict): Data from TransformationDAG selection. These results are
            ignored; data needs to be passed via the result placeholders!
            See above.
        show_hints (bool): Whether to show hints in the case of not passing
            any arguments to a plot function.
        **shared_kwargs (dict): Shared kwargs for all plot functions.
            They are recursively updated, if to_plot specifies the same
            kwargs.

    .. warning::

        Note that especially seaborn plot functions require a ``data``
        argument that needs to be passed via a ``!dag_result`` key,
        see :ref:`dag_result_placeholder`.
        The multiplot function neither expects nor automatically passes a
        ``data`` DAG-node to the individual functions.


    .. note::

        On a failing plot function call the logger will emit a warning.
        This allows to still show the plots of other functions applied on the
        same axis.

    Raises:
        NotImplementedError: On a dict-like ``to_plot`` argument that would
            define the ax to plot on in case of multiple axes to select from.
        TypeError: On a non-list-like or non-dict-like ``to_plot`` argument.
    """
    # dict-like to_plot is not yet implemented
    if isinstance(to_plot, dict):
        raise NotImplementedError(
            "'to_plot' needs to be list-like but was "
            f"of type {type(to_plot)}. Specifying multi-axis plots through "
            "a dict-like 'to_plot' argument is not yet implemented."
        )

    # to_plot needs to be a list
    elif not isinstance(to_plot, (list, tuple)):
        raise TypeError(
            "'to_plot' needs to be list-like but was "
            f"of type {type(to_plot)}. Please assure to pass a list."
        )

    if show_hints and data:
        log.caution(
            "Got the following transformation results via the "
            f"`data` argument: {', '.join(data)}. "
            "Note that the multiplot function ignores these; pass "
            "them via result placeholders instead. Remove these tags "
            "from the `compute_only` argument to avoid passing them "
            "or set `show_hints` to False to suppress this hint."
        )

    for func_num, func_kwargs in enumerate(to_plot):
        # Get the function name, the function object and all function kwargs
        # from the configuration entry.
        func_name, func, func_args, func_kwargs = _parse_func_kwargs(
            shared_kwargs=shared_kwargs, **func_kwargs
        )

        # Notify user if plot functions do not get any kwargs passed on.
        # This is e.g. helpful and relevant for seaborn functions that require
        # a 'data' kwarg but do not fail or warn if no 'data' is passed on to
        # them.
        if (
            show_hints
            and not func_kwargs
            and func_name in _MULTIPLOT_CAUTION_FUNC_NAMES
        ):
            log.caution(
                "Oops, you seem to have called '%s' without any function "
                "arguments. If the plot produces unexpected output, check "
                "that all required arguments (e.g. `data`, `x`, ...) were "
                "given.\n"
                "To silence this warning, set `show_hints` to `False`."
            )

        # Apply the plot function and allow it to fail to make sure that
        # potential other plots are still plotted and shown.
        try:
            func(*func_args, **func_kwargs)

        except Exception as exc:
            msg = (
                f"Plotting with '{func_name}', plot number {func_num}, "
                f"did not succeed! Got a {type(exc).__name__}: {exc}"
            )
            if hlpr.raise_on_error:
                raise PlottingError(msg) from exc
            log.warning(
                f"{msg}\nEnable debug mode to get a full traceback. "
                "Proceeding with next plot ..."
            )



"""This submodule contains some utility classes and functions"""

from .coords import (
    extract_coords,
    extract_coords_from_attrs,
    extract_coords_from_data,
    extract_coords_from_name,
    extract_dim_names,
)
from .data_ops import apply_operation, available_operations, register_operation
from .link import Link
from .ordereddict import IntOrderedDict, KeyOrderedDict



"""Implements common tools of loader functions, including a decorator to
ensure correct loader function signature.
"""


def add_loader(*, TargetCls: type, omit_self: bool = True):
    """This decorator should be used to specify loader functions.

    Args:
        TargetCls: The return type of the load function. This is stored as an
            attribute of the decorated function.
        omit_self (bool, optional): If True (default), the decorated method
            will not be supplied with the ``self`` object instance, thus being
            equivalent to a class method.
    """

    def load_func_decorator(func):
        """This decorator sets the load function's `TargetCls` attribute."""

        def load_func(instance, *args, **kwargs):
            """Calls the load function, either with or without ``self``"""
            load_func.__doc__ = func.__doc__
            if omit_self:
                # class method
                return func(*args, **kwargs)
            # regular method
            return func(instance, *args, **kwargs)

        # Set the target class as function attribute
        load_func.TargetCls = TargetCls

        # Carry over the docstring of the decorated function
        load_func.__doc__ = func.__doc__

        return load_func

    return load_func_decorator
